"""This module implements the :py:class:`~utopya.parameter.Parameter` class
which is used when validating model and simulation parameters.
"""

import logging
import operator
from math import isinf as _isinf
from numbers import Number
from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Union

import numpy as np
import paramspace as psp

from .exceptions import ValidationError

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


class Parameter:
    """The parameter class is used when a model parameter needs to be validated
    before commencing the model run. It can hold information on the parameter
    itself as well as its valid range and type and other meta-data.

    Per default, the ``Parameter`` class should be assumed to handle *scalar*
    parameters like numerical values or strings. For validating sequence-like
    parameters, corresponding specializing classes are to be implemeted.
    """

    # fmt: off
    SHORTHAND_MODES: Dict[str, Callable] = {
        # .. numeric ..........................................................
        "is-positive": lambda d, **kws: dict(
            default=d, limits=(0, None), limits_mode="(]", **kws
        ),
        "is-positive-or-zero": lambda d, **kws: dict(
            default=d, limits=(0, None), limits_mode="[]", **kws
        ),
        "is-negative": lambda d, **kws: dict(
            default=d, limits=(None, 0), limits_mode="[)", **kws
        ),
        "is-negative-or-zero": lambda d, **kws: dict(
            default=d, limits=(None, 0), limits_mode="[]", **kws
        ),

        # .. float ............................................................
        "is-probability": lambda d, **kws: dict(
            default=d, limits=(0, 1), dtype=float, **kws
        ),
        "is-in-unit-interval": lambda d, **kws: dict(
            default=d, limits=(0, 1), dtype=float, **kws
        ),

        # .. integer ..........................................................
        "is-int": lambda d, **kws: dict(
            default=d, dtype=int, **kws
        ),
        "is-positive-int": lambda d, **kws: dict(
            default=d, limits=(0, None), limits_mode="()", dtype=int, **kws
        ),
        "is-negative-int": lambda d, **kws: dict(
            default=d, limits=(None, 0), limits_mode="()", dtype=int, **kws
        ),
        "is-unsigned": lambda d, **kws: dict(
            default=d, limits=(0, None), limits_mode="[)", dtype="uint", **kws
        ),

        # .. other types ......................................................
        "is-string": lambda d, **kws: dict(
            default=d, dtype=str, **kws
        ),
        "is-bool": lambda d, **kws: dict(
            default=d, dtype=bool, **kws
        ),
    }
    """Shorthand mode factory functions.
    These are used in the :py:meth:`~utopya.parameter.Parameter.from_shorthand`
    class method to generate a :py:class:`~utopya.parameter.Parameter` object
    more easily.

    Also, :py:mod:`utopya.yaml` registers each of these shorthand modes as a
    YAML constructor for tag ``!<mode>``.
    """
    # fmt: on

    LIMIT_COMPS: Dict[str, Callable] = {
        "[": operator.ge,
        "(": operator.gt,
        "]": operator.le,
        ")": operator.lt,
    }
    """Comparators for the ``limits`` check, depending on ``limits_mode``"""

    LIMIT_MODES: Sequence[str] = ("[]", "()", "[)", "(]")
    """Possible limit modes"""

    yaml_tag: str = "!param"
    """Default YAML tag to use for representing"""

    # .........................................................................

    def __init__(
        self,
        *,
        default: Any,
        name: str = None,
        description: str = None,
        is_any_of: Sequence[Any] = None,
        limits: Tuple[Optional[float], Optional[float]] = None,
        limits_mode: str = "[]",
        dtype: Union[str, type] = None,
    ):
        """Creates a new Parameter object, which holds a default value as well
        as some constraints on the possible values this parameter can assume.

        Args:
            default (Any): the default value of the parameter.
            name (str, optional): the name of the parameter.
            description (str, optional): a description of this parameter or
                its effects.
            is_any_of (Sequence[Any], optional): a sequence of possible values
                this parameter can assume.
                If this parameter is given, ``limits`` cannot be used.
            limits (Tuple[Optional[float], Optional[float]], optional):
                the upper and lower bounds of the parameter (only applicable
                to scalar numerals). If None, the bound is assumed to be
                negative or positive infinity, respectively. Whether boundary
                values are included into the interval is controlled by the
                ``limits_mode`` argument.
                This argument is mutually exclusive with ``is_any_of``!
            limits_mode (str, optional): whether to interpret the limits as an
                open, closed, or semi-closed interval.
                Possible values: ``'[]'`` (closed, default), ``'()'`` (open),
                ``'[)'``, and ``'(]'``.
            dtype (Union[str, type], optional): expected data type of this
                parameter. Accepts all strings that are accepted by
                :py:class:`numpy.dtype`, eg. ``int``, ``float``, ``uint16``,
                ``string``.

        Raises:
            TypeError: On a ``limits`` argument that was not tuple-like or if
                a ``limits`` argument was given but the ``default`` was a
            ValueError: if an invalid ``limits_mode`` is passed, if ``limits``
                and ``is_any_of`` are *both* passed, or if the ``limits``
                argument did not have length 2.
        """
        if limits_mode not in self.LIMIT_MODES:
            raise ValueError(
                f"Unrecognized `limits_mode` argument "
                f"'{limits_mode}'! "
                f"Choose from: {', '.join(self.LIMIT_MODES)}"
            )

        if limits and is_any_of:
            raise ValueError(
                "The arguments `limits` and `is_any_of` cannot "
                "be specified at the same time!"
            )

        if limits is not None:
            # Ensure tuple, correct length, and numerical default value
            if not isinstance(limits, (tuple, list)):
                raise TypeError(
                    "Expected a tuple-like `limits` argument, but "
                    f"got {type(limits).__name__} with "
                    f"value {limits}!"
                )
            limits = tuple(limits)

            if len(limits) != 2:
                raise ValueError(
                    "`limits` argument should be of length 2! "
                    f"Got: {limits}"
                )

            if default is not None and not isinstance(default, Number):
                raise TypeError(
                    "Unable to use the `limits` argument for non-"
                    f"numerical default value '{default}'! Use "
                    "the `is_any_of` argument instead."
                )

        self._default = default
        self._name = name
        self._description = description
        self._limits = limits
        self._limits_mode = limits_mode
        self._is_any_of = tuple(is_any_of) if is_any_of else None
        self._dtype = np.dtype(dtype) if dtype is not None else None
        self._from_shorthand = None

    # .. Magic Methods ........................................................

    def __eq__(self, other) -> bool:
        """Returns True for parameters with equal behavior."""
        if not isinstance(other, type(self)):
            return False
        return all(
            v == other.__dict__[k]
            for k, v in self.__dict__.items()
            if k not in ("_from_shorthand",)
        )

    def __str__(self) -> str:
        info = {
            attr[1:]: getattr(self, attr)
            for attr in ("_name", "_is_any_of", "_dtype", "_limits")
            if getattr(self, attr) is not None
        }
        if info:
            if "limits" in info:
                info["limits_mode"] = repr(self._limits_mode)
            info_str = ", ".join([f"{k}: {v}" for k, v in info.items()])
            return f"<Parameter, default: {self._default}, {info_str}>"
        return f"<Parameter, default: {self._default}>"

    # .. Properties ...........................................................

    @property
    def default(self):
        """The default value for this parameter"""
        return self._default

    @property
    def name(self):
        """The name of this parameter"""
        return self._name

    @property
    def description(self):
        """The description of this parameter"""
        return self._description

    @property
    def limits(self) -> Union[tuple, None]:
        """The limits of this parameter"""
        return self._limits

    @property
    def limits_mode(self) -> str:
        """The mode used when evaluating the limits"""
        return self._limits_mode

    @property
    def is_any_of(self) -> Tuple[Any]:
        """Possible values this parameter may assume"""
        return self._is_any_of

    @property
    def dtype(self) -> Union[np.dtype, None]:
        """The expected data type of this parameter"""
        return self._dtype

    # .........................................................................

    def validate(self, value: Any, *, raise_exc: bool = True) -> bool:
        """Checks whether the given value would be a valid parameter.

        The checks for the corresponding arguments are carried out in the
        following order:

            1. ``is_any_of``
            2. ``dtype``
            3. ``limits``

        The data type is checked according to the numpy type hierarchy, see
        `docs <https://numpy.org/doc/stable/reference/arrays.scalars.html>`__.
        To reduce strictness, the following additional compatibilities are
        taken into account:

            - for unsigned integer ``dtype``, a signed integer-type
              ``value`` is compatible if ``value >= 0``
            - for floating-point ``dtype``, integer-type ``value`` are always
              considered compatible
            - for floating-point ``dtype``, ``value`` of all floating-point-
              types are considered compatible, even if they have a lower
              precision (note the coercion test below, though)

        Additionally, it is checked whether ``value`` is representable as the
        target data type. This is done by coercing ``value`` to ``dtype`` and
        then checking for equality (using np.isclose).

        Args:
            value (Any): The value to test.
            raise_exc (bool, optional): Whether to raise an exception or not.

        Returns:
            bool: Whether or not the given value is a valid parameter.

        Raises:
            ValidationError: If validation failed or is impossible (for
                instance due to ambiguous validity parameters). This error
                message contains further information on why validation failed.
        """

        def was_invalid(msg: str) -> bool:
            """Handles the case where the given value was not valid.
            This helper function raises a ``ValidationError`` if ``raise_exc``
            flag is set, returns False otherwise.
            """
            if raise_exc:
                raise ValidationError(msg)
            return False

        # Check explicitly given permissible values first
        if self.is_any_of and value not in self.is_any_of:
            _valid_opts = ", ".join([repr(e) for e in self.is_any_of])
            return was_invalid(
                f"value {repr(value)} is not permissible. "
                f"Valid options are:  {_valid_opts}"
            )

        # Check the type
        if self.dtype:
            _is_subtype = np.issubdtype(type(value), self.dtype)

            _float_compatible = np.issubdtype(self.dtype, np.floating) and (
                np.issubdtype(type(value), np.integer)
                or np.issubdtype(type(value), np.floating)
            )
            _uint_compatible = (
                np.issubdtype(self.dtype, np.unsignedinteger)
                and np.issubdtype(type(value), np.integer)
                and value >= 0
            )

            if not (_is_subtype or _float_compatible or _uint_compatible):
                return was_invalid(
                    f"required {self.dtype}-compatible type, but got "
                    f"{type(value).__name__} with value {repr(value)}."
                )

            # For numerical values, check that type coercion would not lead to
            # value changes, e.g. due to values not being representable
            if (
                np.issubdtype(self.dtype, np.number)
                and np.issubdtype(type(value), np.number)
                and not np.isnan(value)
            ):
                _coerced_value = self.dtype.type(value)
                _is_equal = _coerced_value == value
                _is_close = np.isclose(_coerced_value, value)
                # NOTE This is done instead of np.can_cast, which would look
                #      only at the types and thus be too strict. By taking
                #      into account the actual value, we are more compatible.

                if not (_is_equal or _is_close):
                    return was_invalid(
                        f"the {type(value).__name__} value {repr(value)} "
                        "cannot be correctly represented as the required "
                        f"type, {self.dtype}, but would be coerced to "
                        f"{repr(_coerced_value)} != {repr(value)}"
                    )

        # Check value is within the given limits
        if self.limits:
            # Type should be numerical
            if not isinstance(value, Number):
                return was_invalid(
                    f"required numerical value in order to validate limits, "
                    f"but got {type(value).__name__} with value {repr(value)}."
                )

            # Can now assume that it is numerical. Change Nones to ±inf
            lims = list(self.limits)
            if lims[0] is None:
                lims[0] = -float("inf")
            if lims[1] is None:
                lims[1] = float("inf")

            # Get comparison operators according to limits mode
            comp_l = self.LIMIT_COMPS[self.limits_mode[0]]
            comp_r = self.LIMIT_COMPS[self.limits_mode[1]]

            if not (comp_l(value, lims[0]) and comp_r(value, lims[1])):
                # For ±inf bounds, adjust the shown interval to "open"
                return was_invalid(
                    "required value in interval "
                    f"{self.limits_mode[0] if not _isinf(lims[0]) else '('}"
                    f"{lims[0]}, {lims[1]}"
                    f"{self.limits_mode[1] if not _isinf(lims[1]) else ')'}, "
                    f"but got {repr(value)}."
                )

        # If this point is reached, all checks passed: Parameter is valid.
        return True

    # .. Class methods ........................................................

    @classmethod
    def from_shorthand(cls, default: Any, *, mode: str, **kwargs):
        r"""Constructs a Parameter object from a given shorthand mode.

        Args:
            default (Any): the default value for the parameter
            mode (str): A valid shorthand mode, see
                :py:attr:`~utopya.parameter.Parameter.SHORTHAND_MODES`
            **kwargs: any further arguments for Parameter ininitialization,
                see :py:meth:`~utopya.parameter.Parameter.__init__`.

        Returns:
            a Parameter object
        """
        try:
            d = cls.SHORTHAND_MODES[mode](default, **kwargs)

        except KeyError as err:
            raise ValueError(
                f"Unrecognized shorthand mode '{mode}'! "
                f"Needs to be one of: {', '.join(cls.SHORTHAND_MODES)}"
            ) from err

        p = cls(**d)
        p._from_shorthand = dict(default=default, mode=mode, kwargs=kwargs)
        return p

    @classmethod
    def to_yaml(cls, representer, node):
        """Represent this Parameter object as a YAML mapping.

        Args:
            representer (ruamel.yaml.representer): The representer module
            node (Parameter): The node, i.e. an instance of this class

        Returns:
            a yaml mapping that is able to recreate this object
        """
        if node._from_shorthand and not node._from_shorthand.get("kwargs"):
            default = node._from_shorthand["default"]
            mode = node._from_shorthand["mode"]
            return representer.represent_scalar(f"!{mode}", str(default))

        # Not constructed from shorthand
        d = {}
        d["default"] = node.default
        if node.limits:
            d["limits"] = node.limits
            d["limits_mode"] = node.limits_mode
        elif node.is_any_of:
            d["is_any_of"] = node.is_any_of
        if node.dtype:
            d["dtype"] = str(node.dtype)
        return representer.represent_mapping(cls.yaml_tag, d)

    @classmethod
    def from_yaml(cls, constructor, node):
        """The default constructor for Parameter objects, expecting a YAML
        node that is mapping-like.
        """
        return cls(**constructor.construct_mapping(node, deep=True))


# -----------------------------------------------------------------------------


def extract_validation_objects(
    model_cfg: dict, *, model_name: str
) -> Tuple[dict, dict]:
    """Extracts all :py:class:`~utopya.parameter.Parameter` objects from a
    model configuration (a nested dict), replacing them with their default
    values.
    Returns both the modified model configuration well as the Parameter
    objects (keyed by the key sequence necessary to reach them within the
    model configuration).

    Args:
        model_cfg (dict): the model configuration to inspect
        model_name (str): the name of the model

    Returns:
        Tuple[dict, dict]: a tuple of (model config, parameters to validate).
            The model config contains the passed config dict in which all
            Parameter class elements have been replaced by their default
            entries.
            The second entry is a dictionary consisting of the Parameter class
            objects (requiring validation) with keys being key sequences to
            those Parameter objects. Note that the key sequence is relative to
            the level *above* the model configuration, with ``model_name`` as
            a common entry for all returned values.
    """
    # Collect Parameter objects
    is_Parameter_obj = lambda e: isinstance(e, Parameter)
    to_validate = psp.tools.recursive_collect(
        model_cfg, select_func=is_Parameter_obj, prepend_info=("keys",)
    )

    # Replace by default values
    set_default_val = lambda e: e.default
    model_cfg = psp.tools.recursive_replace(
        model_cfg, select_func=is_Parameter_obj, replace_func=set_default_val
    )

    # Prepend the model name to the key sequence such that the key sequences
    # attach *one level above* the model configuration, i.e. at the level of
    # the `parameter_space`.
    to_validate = {
        (model_name,) + ks: v for ks, v in dict(to_validate).items()
    }

    return model_cfg, to_validate



"""Helper module that contains tools useful for module imports or manipulation
of the system path.
These are not implemented here but in :py:mod:`dantro._import_tools`.

.. deprecated:: 1.0.1

    This module is deprecated. Use :py:mod:`dantro._import_tools` or
    :py:func:`utopya_backend.tools.import_package_from_dir` instead.
"""

import warnings

from dantro._import_tools import *

warnings.warn(
    "The utopya._import_tools module has been deprecated and will be removed. "
    "Have a look at `dantro._import_tools` for alternatives. You can also use "
    "the `utopya_backend.tools.import_package_from_dir` function to import "
    "a locally accessible package.",
    DeprecationWarning,
)



"""Test the command line methods"""

import argparse  # FIXME now using click

import pytest

from .._fixtures import *

# import utopya.cltools as clt  # FIXME


# -----------------------------------------------------------------------------


@pytest.mark.skip(reason="needs to be adapted")
def test_prompt_for_new_plot_args(capsys, monkeypatch):
    """Tests the prompt for new plot arguments"""
    # Some mock definition of the parser
    p_eval = argparse.ArgumentParser()
    p_eval.add_argument("model_name")
    p_eval.add_argument("run_dir_path", default=None, nargs="?")

    # plot configuration
    p_eval.add_argument("--plots-cfg", default=None, nargs="?")
    p_eval.add_argument("--update-plots-cfg", default=None, nargs="+")
    p_eval.add_argument("--plot-only", default=None, nargs="*")
    p_eval.add_argument("--interactive", action="store_true")
    p_eval.add_argument("--debug", action="store_true")

    p_eval.add_argument("--run-cfg-path", default=None, nargs="?")
    p_eval.add_argument("--set-cfg", default=None, nargs="+", type=str)
    p_eval.add_argument("--suppress-data-tree", action="store_true")
    p_eval.add_argument("--full-data-tree", action="store_true")
    p_eval.add_argument("--cluster-mode", action="store_true")
    # .........................................................................

    # The prompt function
    prompt = lambda **kws: clt.prompt_for_new_plot_args(**kws, parser=p_eval)

    # Specify mock argv and args
    argv = ["dummy", "--interactive", "--plot-only", "foo", "bar"]
    args = p_eval.parse_args(argv)

    # Specify some example parameters
    monkeypatch.setattr("builtins.input", lambda x: "--plot-only foo")
    argv, args = prompt(old_argv=argv, old_args=args)
    assert "--plot-only" in argv
    assert args.plot_only == ["foo"]
    assert not args.debug

    # Again, with some other input
    # NOTE Can't test the readline startup hook using monkeypatch.setattr, thus
    #      the full user input needs be specified ...
    monkeypatch.setattr(
        "builtins.input", lambda x: "--plot-only foo bar --debug"
    )
    argv, args = prompt(old_argv=argv, old_args=args)
    assert "--plot-only" in argv
    assert "--debug" in argv
    assert args.plot_only == ["foo", "bar"]
    assert args.debug

    # Some bad arguments (that are caught by the parser)
    with pytest.raises(SystemExit):
        monkeypatch.setattr("builtins.input", lambda x: "some pos. argument")
        prompt(old_argv=argv, old_args=args)

    # Some disallowed arguments (that are caught by the prompt function)
    with pytest.raises(ValueError, match="Disallowed arguments:"):
        monkeypatch.setattr(
            "builtins.input", lambda x: "some/run/cfg/path --plot-only foo"
        )
        prompt(old_argv=argv, old_args=args)



"""Implementation of the :py:class:`~utopya.multiverse.Multiverse` class which
sits at the heart of utopya and supplies the main user interface for the
frontend. It allows to run a simulation and then evaluate it.
"""
import copy
import itertools
import logging
import os
import re
import time
import warnings
from collections import defaultdict
from shutil import copy2
from tempfile import TemporaryDirectory
from typing import List, Tuple, Union

import paramspace as psp
from pkg_resources import resource_filename

from ._cluster import parse_node_list
from .cfg import get_cfg_path as _get_cfg_path
from .eval import DataManager, PlotManager
from .model_registry import ModelInfoBundle, get_info_bundle, load_model_cfg
from .parameter import ValidationError
from .project_registry import PROJECTS
from .reporter import WorkerManagerReporter
from .tools import parse_num_steps, pformat, recursive_update
from .workermanager import WorkerManager
from .yaml import load_yml, write_yml

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class Multiverse:
    """The Multiverse is where a single simulation run is orchestrated from.

    It spawns multiple universes, each of which represents a single simulation
    of the selected model with the parameters specified by the meta
    configuration.
    The :py:class:`~utopya.workermanager.WorkerManager` takes care to perform
    these simulations in parallel.

    The :py:class:`.Multiverse` then interfaces with the :py:mod:`dantro` data
    processing pipeline using classes specialized in :py:mod:`utopya.eval`:
    The :py:class:`~utopya.eval.datamanager.DataManager` loads the created
    simulation output, making it available in a uniformly accessible and
    hierarchical data tree.
    Then, the :py:class:`~utopya.eval.plotmanager.PlotManager` handles
    plotting of that data.
    """

    BASE_META_CFG_PATH = resource_filename("utopya", "cfg/base_cfg.yml")
    """Where the default meta configuration can be found"""

    USER_CFG_SEARCH_PATH = _get_cfg_path("user")
    """Where to look for the user configuration"""

    RUN_DIR_TIME_FSTR = "%y%m%d-%H%M%S"
    """The time format string for the run directory"""

    UTOPYA_BASE_PLOTS_PATH = resource_filename("utopya", "cfg/base_plots.yml")
    """Where the utopya base plots configuration can be found; this is passed
    to the :py:class:`~utopya.eval.plotmanager.PlotManager`.
    """

    def __init__(
        self,
        *,
        model_name: str = None,
        info_bundle: ModelInfoBundle = None,
        run_cfg_path: str = None,
        user_cfg_path: str = None,
        _shared_worker_manager: WorkerManager = None,
        **update_meta_cfg,
    ):
        """Initialize the Multiverse.

        Args:
            model_name (str, optional): The name of the model to run
            info_bundle (ModelInfoBundle, optional): The model information
                bundle that includes information about the executable path etc.
                If not given, will attempt to read it from the model registry.
            run_cfg_path (str, optional): The path to the run configuration.
            user_cfg_path (str, optional): If given, this is used to update the
                base configuration. If None, will look for it in the default
                path, see Multiverse.USER_CFG_SEARCH_PATH.
            _shared_worker_manager (WorkerManager, optional): If given, this
                already existing WorkerManager instance (and its reporter)
                will be used instead of initializing new instances.

                .. warning::

                    This argument is only exposed for internal purposes.
                    It should not be used for production code and behavior of
                    this argument may change at any time.

            **update_meta_cfg: Can be used to update the meta configuration
                generated from the previous configuration levels
        """
        # First things first: get the info bundle
        if info_bundle is None:
            info_bundle = get_info_bundle(
                model_name=model_name, info_bundle=info_bundle
            )
        self._info_bundle = info_bundle

        log.progress(
            "Initializing Multiverse for '%s' model ...", self.model_name
        )

        # Setup property-managed attributes
        self._dirs = dict()
        self._model_executable = None
        self._tmpdir = None
        self._resolved_cluster_params = None

        # Create meta configuration and list of used config files
        mcfg, cfg_parts = self._create_meta_cfg(
            run_cfg_path=run_cfg_path,
            user_cfg_path=user_cfg_path,
            update_meta_cfg=update_meta_cfg,
        )
        self._meta_cfg = mcfg
        log.info("Built meta configuration.")
        log.remark("  Debug level:  %d", self.debug_level)
        self._apply_debug_level()

        # In cluster mode, need to make some adjustments via additional dicts
        dm_cluster_kwargs = dict()
        wm_cluster_kwargs = dict()
        if self.cluster_mode:
            log.note("Cluster mode enabled.")
            self._resolved_cluster_params = self._resolve_cluster_params()
            rcps = self.resolved_cluster_params  # creates a deep copy

            log.note(
                "This is node %d of %d.",
                rcps["node_index"] + 1,
                rcps["num_nodes"],
            )

            # Changes to the meta configuration
            # To avoid config file collisions in the PlotManager:
            self._meta_cfg["plot_manager"]["cfg_exists_action"] = "skip"

            # _Additional_ arguments to pass to *Manager initializations below
            # ... for DataManager
            timestamp = rcps["timestamp"]
            dm_cluster_kwargs = dict(
                out_dir_kwargs=dict(timestamp=timestamp, exist_ok=True)
            )

            # ... for WorkerManager
            wm_cluster_kwargs = dict(
                cluster_mode=True, resolved_cluster_params=rcps
            )

        # Create the run directory and write the meta configuration into it.
        self._create_run_dir(**self.meta_cfg["paths"])
        log.note("Run directory:\n  %s", self.dirs["run"])

        # Backup involved files, if not in cluster mode or on the relevant node
        if (
            not self.cluster_mode
            or self.resolved_cluster_params["node_index"] == 0
        ):
            # If not in cluster mode, should backup in any case.
            # In cluster mode, the first node is responsible for backing up
            # the configuration; all others can relax.
            self._perform_backup(
                **self.meta_cfg["backups"], cfg_parts=cfg_parts
            )

        else:
            log.debug(
                "Not backing up config files, because it was already "
                "taken care of by the first node."
            )
            # NOTE Not taking a try-except approach here because it might get
            #      messy when multiple nodes try to backup the configuration
            #      at the same time ...

        # Validate the parameters specified in the meta configuration
        self._validate_meta_cfg()

        # Prepare the executable
        self._prepare_executable(**self.meta_cfg["executable_control"])

        # Create a DataManager instance
        self._dm = DataManager(
            self.dirs["run"],
            name=f"{self.model_name}_data",
            **self.meta_cfg["data_manager"],
            **dm_cluster_kwargs,
        )
        log.progress("Initialized DataManager.")

        # Either create a WorkerManager instance and its associated reporter
        # or use an already existing WorkerManager that is also used elsewhere
        if not _shared_worker_manager:
            self._wm = WorkerManager(
                **self.meta_cfg["worker_manager"], **wm_cluster_kwargs
            )
            self._reporter = WorkerManagerReporter(
                self.wm,
                mv=self,
                report_dir=self.dirs["run"],
                **self.meta_cfg["reporter"],
            )
        else:
            self._wm = _shared_worker_manager
            self._reporter = self.wm.reporter
            log.info("Using a shared WorkerManager instance and reporter.")

        # And instantiate the PlotManager with the model-specific plot config
        self._pm = self._setup_pm()

        log.progress("Initialized Multiverse.\n")

    # Properties ..............................................................

    @property
    def debug_level(self) -> int:
        """The debug level"""
        return self.meta_cfg.get("debug_level", 0)

    @property
    def info_bundle(self) -> ModelInfoBundle:
        """The model info bundle for this Multiverse"""
        return self._info_bundle

    @property
    def model_name(self) -> str:
        """The model name associated with this Multiverse"""
        return self.info_bundle.model_name

    @property
    def model_executable(self) -> str:
        """The path to the model executable"""
        if self._model_executable is not None:
            # Use the executable from a temporary directory
            return self._model_executable

        execpath = self.info_bundle.executable
        if not execpath:
            raise ValueError(
                f"Model '{self.model_name}' does not have an executable "
                "registered that can be used to perform a simulation run!\n"
                "If you want to run simulations, specify an executable. "
                "In case you only want to use utopya's evaluation routines, "
                "no executable is needed, but only the evaluation pipeline is "
                "available and it seems like you tried to run a simulation."
            )

        return execpath

    @property
    def model(self) -> "utopya.model.Model":
        """A model instance, created ad-hoc using the associated info bundle"""
        from .model import Model

        return Model(info_bundle=self.info_bundle)

    @property
    def meta_cfg(self) -> dict:
        """The meta configuration."""
        return self._meta_cfg

    @property
    def dirs(self) -> dict:
        """Information on managed directories."""
        return self._dirs

    @property
    def cluster_mode(self) -> bool:
        """Whether the Multiverse should run in cluster mode"""
        return self.meta_cfg["cluster_mode"]

    @property
    def cluster_params(self) -> dict:
        """Returns a copy of the cluster mode configuration parameters"""
        return copy.deepcopy(self.meta_cfg["cluster_params"])

    @property
    def resolved_cluster_params(self) -> dict:
        """Returns a copy of the cluster configuration with all parameters
        resolved. This makes some additional keys available on the top level.
        """
        # Return the cached value as a _copy_ to secure it against changes
        return copy.deepcopy(self._resolved_cluster_params)

    @property
    def dm(self) -> DataManager:
        """The Multiverse's DataManager."""
        return self._dm

    @property
    def wm(self) -> WorkerManager:
        """The Multiverse's WorkerManager."""
        return self._wm

    @property
    def pm(self) -> PlotManager:
        """The Multiverse's PlotManager."""
        return self._pm

    # Public methods ..........................................................

    def run(self, *, sweep: bool = None):
        """Starts a Utopia simulation run.

        Specifically, this method adds simulation tasks to the associated
        WorkerManager, locks its task list, and then invokes the
        :py:meth:`~utopya.workermanager.WorkerManager.start_working` method
        which performs all the simulation tasks.

        If cluster mode is enabled, this will split up the parameter space into
        (ideally) equally sized parts and only run one of these parts,
        depending on the cluster node this Multiverse is being invoked on.

        .. note::

            As this method locks the task list of the
            :py:class:`~utopya.workermanager.WorkerManager`, no further tasks
            can be added henceforth. This means, that each Multiverse instance
            can only perform a single simulation run.

        Args:
            sweep (bool, optional): Whether to perform a sweep or not. If None,
                the value will be read from the ``perform_sweep`` key of the
                meta-configuration.
        """
        log.info("Preparing for simulation run ...")

        # Add tasks, then prevent adding further tasks to disallow further runs
        self._add_sim_tasks(sweep=sweep)
        self.wm.tasks.lock()

        # Tell the WorkerManager to start working (is a blocking call)
        self.wm.start_working(**self.meta_cfg["run_kwargs"])

        # Done! :)
        log.success("Finished simulation run. Wohoo. :)\n")

    def run_single(self):
        """Runs a single simulation using the parameter space's default value.

        See :py:meth:`~utopya.multiverse.Multiverse.run` for more information.
        """
        return self.run(sweep=False)

    def run_sweep(self):
        """Runs a parameter sweep.

        See :py:meth:`~utopya.multiverse.Multiverse.run` for more information.
        """
        return self.run(sweep=True)

    def renew_plot_manager(self, **update_kwargs):
        """Tries to set up a new PlotManager. If this succeeds, the old one is
        discarded and the new one is associated with this Multiverse.

        Args:
            **update_kwargs: Passed on to PlotManager.__init__
        """
        try:
            pm = self._setup_pm(**update_kwargs)

        except Exception as exc:
            raise ValueError(
                "Failed setting up a new PlotManager! "
                "The old PlotManager remains."
            ) from exc

        self._pm = pm

    # Helpers .................................................................

    def _create_meta_cfg(
        self, *, run_cfg_path: str, user_cfg_path: str, update_meta_cfg: dict
    ) -> dict:
        """Create the meta configuration from several parts and store it.

        The final configuration dict is built from multiple components,
        where one is always recursively updating the previous level.
        The resulting configuration is the meta configuration and is stored
        to the ``meta_cfg`` attribute.

        The parts are recorded in the ``cfg_parts`` dict and returned such that
        a backup can be created.

        Args:
            run_cfg_path (str): path to the run configuration
            user_cfg_path (str): path to the user configuration file
            update_meta_cfg (dict): will be used to update the resulting dict

        Returns:
            dict: dict of the parts that were needed to create the meta config.
                The dict-key corresponds to the part name, the value is the
                payload which can be either a path to a cfg file or a dict
        """
        log.info("Building meta-configuration ...")

        # Read in the base meta configuration
        base_cfg = load_yml(self.BASE_META_CFG_PATH)

        # Framework- and project-level configuration files
        framework_cfg = {}
        framework_cfg_path = None

        project_cfg = {}
        project_cfg_path = None

        model_mv_cfg = {}
        model_mv_cfg_path = None

        project_name = self.info_bundle.project_name
        if project_name:
            project = self.info_bundle.project

            # Framework-level
            framework_name = project.get("framework_name")
            if framework_name:
                framework_project = PROJECTS[framework_name]
                framework_cfg_path = framework_project["paths"].get(
                    "mv_project_cfg"
                )

                if framework_cfg_path:
                    framework_cfg = load_yml(framework_cfg_path)

            # Project level
            if project_name != framework_name:
                project_cfg_path = self.info_bundle.project["paths"].get(
                    "mv_project_cfg"
                )

                if project_cfg_path:
                    project_cfg = load_yml(project_cfg_path)

        # There might be a model-specific configuration
        if model_mv_cfg_path := self.info_bundle.paths.get("mv_model_cfg"):
            model_mv_cfg = load_yml(model_mv_cfg_path)

        # Decide whether to read in the user configuration from the default
        # search location or use a user-passed one
        if user_cfg_path is None:
            log.debug(
                "Looking for user configuration file in default "
                "location, %s",
                self.USER_CFG_SEARCH_PATH,
            )

            if os.path.isfile(self.USER_CFG_SEARCH_PATH):
                user_cfg_path = self.USER_CFG_SEARCH_PATH
            else:
                # No user cfg will be loaded
                log.debug("No file found at the default search location.")

        elif user_cfg_path is False:
            log.debug(
                "Not loading the user configuration from the default "
                "search path: %s",
                self.USER_CFG_SEARCH_PATH,
            )

        user_cfg = None
        if user_cfg_path:
            user_cfg = load_yml(user_cfg_path)

        # Read in the configuration corresponding to the chosen model
        (model_cfg, model_cfg_path, params_to_validate) = load_model_cfg(
            info_bundle=self.info_bundle
        )
        # NOTE Unlike the other configuration files, this does not attach at
        # root level of the meta configuration but parameter_space.<model_name>
        # in order to allow it to be used as the default configuration for an
        # _instance_ of that model.

        # Read in the run configuration
        run_cfg = None
        if run_cfg_path:
            log.note("Run configuration:\n  %s", run_cfg_path)
            try:
                run_cfg = load_yml(run_cfg_path)
            except FileNotFoundError as err:
                raise FileNotFoundError(
                    f"No run config could be found at {run_cfg_path}!"
                ) from err

            # Make sure its parameter_space entry is a true dict and not a
            # ParamSpace object, which needs to be resolved to a dict.
            if isinstance(run_cfg.get("parameter_space"), psp.ParamSpace):
                run_cfg["parameter_space"] = run_cfg["parameter_space"]._dict
                # FIXME Should not be using the private interface here. This is
                #       only necessary in the first place because ParamSpace
                #       does not behave like a dict as much as it should...
                #       Once recursive_update works within ParamSpace, this
                #       can be removed again.

        else:
            log.note(
                "Using default run configuration:\n  %s\n", model_cfg_path
            )
        # After this point it is assumed that all values are valid.
        # Those keys or values will throw errors once they are used ...

        # Now perform the recursive update steps
        # Start with the base configuration
        meta_tmp = base_cfg

        # Update with framework and project config, if available
        if framework_cfg:
            log.debug("Updating with framework configuration ...")
            meta_tmp = recursive_update(meta_tmp, framework_cfg)

        if project_cfg:
            log.debug("Updating with project configuration ...")
            meta_tmp = recursive_update(meta_tmp, project_cfg)

        if model_mv_cfg:
            log.debug("Updating with model-specific configuration ...")
            meta_tmp = recursive_update(meta_tmp, model_mv_cfg)

        # Update with user configuration, if given
        if user_cfg:
            log.debug("Updating with user configuration ...")
            meta_tmp = recursive_update(meta_tmp, user_cfg)

        # In order to incorporate the model config, the parameter space is
        # needed. We can already be sure that the parameter_space key exists,
        # because it is added as part of the base_cfg.
        pspace = meta_tmp["parameter_space"]

        # Adjust parameter space to include model configuration at a specified
        # key; also communicate that key explicitly.
        log.debug(
            "Updating parameter space with default model configuration for "
            "model '%s' ...",
            self.model_name,
        )
        pspace[self.model_name] = recursive_update(
            pspace.get(self.model_name, {}), model_cfg
        )
        pspace["root_model_name"] = self.model_name
        # NOTE this works because meta_tmp is a dict and thus mutable :)

        # On top of all of that: add the run configuration, if given
        if run_cfg:
            log.debug("Updating with run configuration ...")
            meta_tmp = recursive_update(meta_tmp, run_cfg)

        # ... and the update_meta_cfg dictionary
        if update_meta_cfg:
            log.debug("Updating with given `update_meta_cfg` dictionary ...")
            meta_tmp = recursive_update(
                meta_tmp, copy.deepcopy(update_meta_cfg)
            )
            # NOTE using deep copy to make sure that usage of the dict will not
            #      interfere with the Multiverse's meta config

        # Make `parameter_space` a ParamSpace object
        pspace = meta_tmp["parameter_space"]
        meta_tmp["parameter_space"] = psp.ParamSpace(pspace)
        log.debug("Converted parameter_space to ParamSpace object.")

        # Add the parameters that require validation
        meta_tmp["parameters_to_validate"] = params_to_validate
        log.debug(
            "Added %d parameters requiring validation.",
            len(params_to_validate),
        )

        # Prepare dict to store paths for config files in (for later backup)
        log.debug("Preparing dict of config parts ...")
        cfg_parts = dict(
            base=self.BASE_META_CFG_PATH,
            framework=framework_cfg_path,
            project=project_cfg_path,
            model_mv=model_mv_cfg,
            user=user_cfg_path,
            model=model_cfg_path,
            run=run_cfg_path,
            update=update_meta_cfg,
        )
        return meta_tmp, cfg_parts

    def _apply_debug_level(self, lvl: int = None):
        """Depending on the debug level, applies certain settings to the
        Multiverse and the runtime environment.

        .. note::

            This does *not* (yet) set the corresponding debug flags for the
            ``PlotManager``, ``DataManager``, or ``WorkerManager``!
        """
        lvl = lvl if lvl is not None else self.debug_level

        if lvl >= 2:
            warnings.simplefilter("always", DeprecationWarning)

    def _create_run_dir(
        self,
        *,
        out_dir: str,
        model_note: str = None,
        dir_permissions: dict = None,
    ) -> None:
        """Create the folder structure for the run output.

        For the chosen model name and current timestamp, the run directory
        will be of form <timestamp>_<model_note> and be part of the following
        directory tree:

        ::

            utopya_output
                model_a
                    180301-125410_my_model_note
                        config
                        data
                            uni000
                            uni001
                            ...
                        eval
                model_b
                    180301-125412_my_first_sim
                    180301-125413_my_second_sim

        If running in cluster mode, the cluster parameters are resolved and
        used to determine the name of the simulation. The pattern then does not
        include a timestamp as each node might return not quite the same value.
        Instead, a value from an environment variable is used.
        The resulting path can have different forms, depending on which
        environment variables were present; required parts are denoted by a
        ``*`` in the following pattern; if the value of the other entries is
        not available, the connecting underscore will not be used:

        ::

            {timestamp}_{job id*}_{cluster}_{job account}_{job name}_{note}

        Args:
            out_dir (str): The base output directory, where all Utopia output
                is stored.
            model_note (str, optional): The note to add to the run directory
                of the current run.
            dir_permissions (Dict[str, str]): If given, will set directory
                permissions on the specified managed directories of this
                Multiverse. The keys of this dict should be entries of the
                :py:attr:`.dirs` attribute, values should be octal permissions
                values given as a string.

        Raises:
            RuntimeError: If the simulation directory already existed. This
                should not occur, as the timestamp is unique. If it occurs,
                you either started two simulations very close to each other or
                something is seriously wrong. Strange time zone perhaps?
        """
        # Define a list of format string parts, starting with timestamp
        fstr_parts = ["{timestamp:}"]

        # Add respective information, depending on mode
        if not self.cluster_mode:
            # Available information is only the timestamp and the model note
            fstr_kwargs = dict(
                timestamp=time.strftime(self.RUN_DIR_TIME_FSTR),
                model_note=model_note,
            )

        else:
            # In cluster mode, need to resolve cluster parameters first
            rcps = self.resolved_cluster_params

            # Now, gather all information for the format string that will
            # determine the name of the output directory. Make all the info
            # available that was supplied from environment variables
            fstr_kwargs = {
                k: v for k, v in rcps.items() if k not in ("custom_out_dir",)
            }

            # Parse timestamp and model note separately
            timestr = time.strftime(
                self.RUN_DIR_TIME_FSTR, time.gmtime(rcps["timestamp"])
            )
            fstr_kwargs["timestamp"] = timestr  # overwrites existing
            fstr_kwargs["model_note"] = model_note  # may be None

            # Add the additional run dir format string parts; its the user's
            # responsibility to supply something reasonable here.
            if self.cluster_params.get("additional_run_dir_fstrs"):
                fstr_parts += self.cluster_params["additional_run_dir_fstrs"]

            # Now, also allow a custom output directory
            if rcps.get("custom_out_dir"):
                out_dir = rcps["custom_out_dir"]

        # Have the model note as suffix
        if model_note:
            fstr_parts += ["{model_note:}"]

        # fstr_parts and fstr_kwargs ready now. Carry out the format operation.
        fstr = "_".join(fstr_parts)
        run_dir_name = fstr.format(**fstr_kwargs)
        log.debug("Determined run directory name:  %s", run_dir_name)

        # Parse the output directory, then build the run directory path
        log.debug("Creating path for run directory inside %s ...", out_dir)
        out_dir = os.path.expanduser(str(out_dir))

        run_dir = os.path.join(out_dir, self.model_name, run_dir_name)
        log.debug("Built run directory path:  %s", run_dir)
        self.dirs["run"] = run_dir

        # ... and create it. In cluster mode, it may already exist.
        try:
            os.makedirs(run_dir, exist_ok=self.cluster_mode)

        except OSError as err:
            raise RuntimeError(
                "Simulation directory already exists. This "
                "should not have happened and is probably due "
                "to two simulations having been started at "
                "almost the same time. Try to start the "
                "simulation again or add a unique model note."
            ) from err

        log.debug("Created run directory.")

        # Create the subfolders that are always assumed to be present
        for subdir in ("config", "data", "eval"):
            subdir_path = os.path.join(run_dir, subdir)
            os.makedirs(subdir_path, exist_ok=self.cluster_mode)
            self.dirs[subdir] = subdir_path

        log.debug("Created subdirectories:  %s", self._dirs)

        # May want to adapt directory permissions
        if not dir_permissions:
            return

        for dirname, mode in dir_permissions.items():
            if mode is None:
                continue
            mode = int(str(mode), 8)
            log.debug(
                "Setting permissions on %s directory to %s ...",
                dirname,
                oct(mode),
            )
            os.chmod(self.dirs[dirname], mode)

    def _setup_pm(self, **update_kwargs) -> PlotManager:
        """Helper function to setup a PlotManager instance"""
        pm_kwargs = copy.deepcopy(self.meta_cfg["plot_manager"])

        if update_kwargs:
            pm_kwargs = recursive_update(pm_kwargs, update_kwargs)

        base_cfg_pools = pm_kwargs.pop(
            "base_cfg_pools", ["utopya_base", "model_base"]
        )

        log.info("Initializing PlotManager ...")
        pm = PlotManager(
            dm=self.dm,
            _model_info_bundle=self.info_bundle,
            default_plots_cfg=self.info_bundle.paths.get("default_plots"),
            base_cfg_pools=self._parse_base_cfg_pools(base_cfg_pools),
            **pm_kwargs,
        )

        log.progress("Initialized PlotManager.")
        log.note(
            "Available base plot configuration pools:\n  %s",
            ", ".join(pm.base_cfg_pools.keys()),
        )
        log.note(
            "Output directory:  %s",
            pm._out_dir if pm._out_dir else "\n  " + self.dm.dirs["out"],
        )

        return pm

    def _parse_base_cfg_pools(
        self, base_cfg_pools: List[Union[str, Tuple[str, Union[str, dict]]]]
    ) -> List[Tuple[str, Union[str, dict]]]:
        """Prepares the ``base_cfg_pools`` argument to be valid input to the
        PlotManager. This method resolves format strings and thus allows to
        more generically define base config pools.

        Possible formats for each entry of ``base_cfg_pools`` argument are:

            - A 2-tuple ``(name, pool dict)`` which specifies the name of the
              base config pool alongside with the pool entries.
            - A 2-tuple ``(name, path to pool config file)``, which is later
              loaded by the PlotManager
            - A shortcut key which resolves to the corresponding 2-tuple.
              Available shortcuts are:  ``utopya_base``, ``framework_base``,
              ``project_base``, and ``model_base``.

        Both the pool name and path may be format strings which get resolved
        with the ``model_name`` key and (in the case of the path) the full
        ``paths`` dict of the current model's info bundle. A format string may
        look like this:

            "{paths[source_dir]}/{model_name}_more_plots.yml"
            "~/some/more/plots/{model_name}/plots.yml"

        If such a path cannot be resolved, an error is logged and an empty pool
        is used instead; this allows for more flexibility in defining locations
        for additional config pools.

        Args:
            base_cfg_pools (List[Union[str, Tuple[str, Union[str, dict]]]]):
                The unparsed specification of config pools.
        """

        def parse_entry(
            entry: Union[str, list, tuple], replacements: dict
        ) -> Tuple[str, Union[str, dict]]:
            """Unpacks an entry into (name, pool) format and resolves any
            remaining format specifiers in the name or pool path.
            """
            if isinstance(entry, str):
                try:
                    pool_name, pool = replacements[entry]
                except KeyError as err:
                    _avail = ", ".join(replacements)
                    raise ValueError(
                        f"Invalid base config pool shortcut key '{entry}'! "
                        f"Available shortcuts are: {_avail}. "
                        "Use one of those or specify the config pool as a "
                        "2-tuple in form (name, path to pool)."
                    ) from err
            else:
                pool_name, pool = entry

            # Parse pool name and the path to the pool config file
            pool_name = pool_name.format(model_name=self.model_name)
            if isinstance(pool, str):
                pool = pool.format(model_name=self.model_name, paths=paths)
                pool = os.path.abspath(os.path.expanduser(pool))
                if not os.path.isfile(pool):
                    log.error(
                        "No base plot config pool file found at:\n  %s", pool
                    )
                    log.caution("Using an empty pool instead.")
                    pool = {}

            # Make sure it is either a dict or a string, not a Path-like object
            if not isinstance(pool, dict):
                pool = str(pool)

            return pool_name, pool

        if not isinstance(base_cfg_pools, list):
            raise TypeError(
                "Base config pools need to be specified as a list of "
                f"2-tuples or strings, got {type(base_cfg_pools)}!"
            )

        paths = self.info_bundle.paths
        replacements = dict(
            utopya_base=("utopya", self.UTOPYA_BASE_PLOTS_PATH),
            framework_base=("framework", {}),
            project_base=("project", {}),
            model_base=("{model_name:}_base", paths.get("base_plots", {})),
        )

        project = self.info_bundle.project
        if project:
            fw_name = project.get("framework_name")
            if fw_name:
                fw_project = PROJECTS[fw_name]
                replacements["framework_base"] = (
                    "framework",
                    fw_project["paths"].get("project_base_plots"),
                )

            # Only add a project-level replacement if it is different from the
            # framework-level replacement
            if self.info_bundle.project_name != fw_name and project[
                "paths"
            ].get("project_base_plots"):
                replacements["project_base"] = (
                    "project",
                    project["paths"].get("project_base_plots"),
                )

        return [parse_entry(p, replacements) for p in base_cfg_pools]

    def _perform_backup(
        self,
        *,
        cfg_parts: dict,
        backup_cfg_files: bool = True,
        backup_executable: bool = False,
        include_git_info: bool = True,
    ) -> None:
        """Performs a backup of that information that can be used to recreate a
        simulation.

        The configuration files are backed up into the ``config`` subdirectory
        of the run directory. All other relevant information is stored in an
        additionally created ``backup`` subdirectory.

        .. warning::

            These backups are created prior to the start of the actual
            simulation run and contains information known at that point.
            Any changes to the meta configuration made *after* initialization
            of the Multiverse will not be reflected in these backups.

            In particular, the ``perform_sweep`` and ``parameter_space``
            entries of the meta configuration may not reflect which form of
            parameter space iteration was actually performed, because the
            ``run_single`` and ``run_sweep`` methods overwrite this behavior.
            To that end, that information is separately stored once the ``run``
            methods are invoked.

        Args:
            cfg_parts (dict): A dict of either paths to configuration files or
                dict-like data that is to be dumped into a configuration file.
            backup_cfg_files (bool, optional): Whether to backup the individual
                configuration files (i.e. the ``cfg_parts`` information). If
                false, the meta configuration will still be backed up.
            backup_executable (bool, optional): Whether to backup the
                executable. Note that these files can sometimes be quite large.
            include_git_info (bool, optional): If True, will store information
                about the state of the project's (and framework's, if existent)
                git repository.
        """
        log.info("Performing backups ...")
        cfg_dir = self.dirs["config"]

        # Write the meta config to the config directory.
        write_yml(self.meta_cfg, path=os.path.join(cfg_dir, "meta_cfg.yml"))
        log.note("  Backed up meta configuration.")

        # Store the *full* parameter space and its metadata
        # NOTE This data may not be equivalent to the parameter space that is
        #      used for a simulation run; another backup is performed when
        #      adding the corresponding simulation tasks.
        _pspace_info = dict(perform_sweep=self.meta_cfg.get("perform_sweep"))
        self._perform_pspace_backup(
            self.meta_cfg["parameter_space"],
            filename="full_parameter_space",
            **_pspace_info,
        )

        # If configured, backup the other cfg files one by one.
        if backup_cfg_files:
            log.debug(
                "Backing up %d involved configuration parts...", len(cfg_parts)
            )

            for part_name, val in cfg_parts.items():
                _path = os.path.join(cfg_dir, part_name + "_cfg.yml")

                # Distinguish two types of payload that will be saved:
                if isinstance(val, str):
                    # Assumed to be path to a config file; copy it
                    log.debug("Copying %s config ...", part_name)
                    copy2(val, _path)

                elif isinstance(val, dict):
                    log.debug("Dumping %s config dict ...", part_name)
                    write_yml(val, path=_path)

            log.note("  Backed up all involved configuration files.")

        # If enabled, back up the executable as well
        if backup_executable:
            backup_dir = os.path.join(self.dirs["run"], "backup")
            os.makedirs(backup_dir, exist_ok=True)

            copy2(
                self.model_executable,
                os.path.join(backup_dir, self.model_name),
            )
            log.note("  Backed up executable.")

        # If enabled, store git repository information
        if include_git_info:
            prj = self.info_bundle.project
            if prj is not None:
                prj_info = prj.get_git_info(include_patch_info=True)

                patch = prj_info.pop("git_diff")
                _patch_path = os.path.join(cfg_dir, "git_diff_project.patch")
                prj_info["git_diff"] = _patch_path

                # write info
                _path = os.path.join(cfg_dir, "git_info_project.yml")
                write_yml(prj_info, path=_path)

                # write patch file
                with open(_patch_path, "w") as f:
                    f.write(patch)

                log.note("  Stored project's git info and patch.")

                fw = prj.framework_project
                if fw is not None:
                    fw_info = fw.get_git_info(include_patch_info=True)

                    patch = fw_info.pop("git_diff")
                    _patch_path = os.path.join(
                        cfg_dir, "git_diff_framework.patch"
                    )
                    fw_info["git_diff"] = _patch_path

                    # write info
                    _path = os.path.join(cfg_dir, "git_info_framework.yml")
                    write_yml(fw_info, path=_path)

                    # write patch file
                    with open(_patch_path, "w") as f:
                        f.write(patch)

                    log.note("  Stored framework's git info and patch.")

    def _perform_pspace_backup(
        self,
        pspace: psp.ParamSpace,
        *,
        filename: str = "parameter_space",
        **info_kwargs,
    ):
        """Stores the given parameter space and its metadata into the
        ``config`` directory.
        Two files will be produced:

            - ``config/{filename}.yml``: the passed ``pspace`` object
            - ``config/{filename}_info.yml``: the passed ``pspace`` object's
                info dictionary containing relevant metadata (and the
                additionally passed ``info_kwargs``)

        .. note::

            This method is separated from the regular backup method
            :py:meth:`._perform_backup` because the
            parameter space that is *used* during a simulation run may be a
            lower-dimensional version of the one the Multiverse was
            initialized with.
            To that end, :py:meth:`.run` will invoke this backup function
            again once the relevant information is fully determined. This is
            important because it is needed to communicate the correct
            information about the sweep to objects downstream in the pipeline
            (e.g. :py:class:`~utopya.eval.plotcreators.MultiversePlotCreator`).

        Args:
            pspace (paramspace.paramspace.ParamSpace): The ParamSpace object
                to save as backup.
            filename (str, optional): The filename (without extension!) to use.
                (This is also used for the log message, with underscores
                changed to spaces.)
            **info_kwargs: Additional kwargs that are to be stored in the meta-
                data dict.
        """
        cfg_dir = self.dirs["config"]
        write_yml(pspace, path=os.path.join(cfg_dir, f"{filename}.yml"))
        write_yml(
            dict(**pspace.get_info_dict(), **info_kwargs),
            path=os.path.join(cfg_dir, f"{filename}_info.yml"),
        )
        log.note("  Backed up %s and metadata.", filename.replace("_", " "))

    def _prepare_executable(self, *, run_from_tmpdir: bool = False) -> None:
        """Prepares the model executable, potentially copying it to a temporary
        location.

        Note that ``run_from_tmpdir`` requires the executable to be relocatable
        to another location, i.e. be position-independent.

        Args:
            run_from_tmpdir (bool, optional): Whether to copy the executable
                to a temporary directory that goes out of scope once the
                Multiverse instance goes out of scope.

        Raises:
            FileNotFoundError: On missing file at model binary path
            PermissionError: On wrong access rights of file at the binary path
        """
        execpath = self.model_executable

        # Make sure it exists and is executable
        if not os.path.isfile(execpath):
            raise FileNotFoundError(
                "No file found at the specified executable "
                f"path for model '{self.model_name}'! "
                "If your model needs building, did you build it?\n"
                f"Expected file at:  {execpath}"
            )

        elif not os.access(execpath, os.X_OK):
            raise PermissionError(
                f"The specified executable path for model '{self.model_name}' "
                "does not point to an executable file. Did you set the "
                f"correct access rights?\n"
                "Use the chmod command to mark the file as executable:\n\n"
                f"  chmod +x {execpath}\n"
            )

        if run_from_tmpdir:
            self._tmpdir = TemporaryDirectory(prefix=self.model_name)
            tmp_execpath = os.path.join(
                self._tmpdir.name, os.path.basename(execpath)
            )

            log.info("Copying executable to temporary directory ...")
            log.debug("  Original:   %s", execpath)
            log.debug("  Temporary:  %s", tmp_execpath)
            copy2(execpath, tmp_execpath)
            execpath = tmp_execpath

        self._model_executable = execpath

    def _resolve_cluster_params(self) -> dict:  # TODO Outsource!
        """This resolves the cluster parameters, e.g. by setting parameters
        depending on certain environment variables. This function is called by
        the resolved_cluster_params property.

        Returns:
            dict: The resolved cluster configuration parameters

        Raises:
            ValueError: If a required environment variable was missing or empty
        """

        log.debug("Resolving cluster parameters from environment ...")

        # Get a copy of the meta configuration parameters
        cps = self.cluster_params

        # Determine the environment to use; defaults to os.environ
        env = cps.get("env") if cps.get("env") else dict(os.environ)

        # Get the mapping of environment variables to target variables
        mngr = cps["manager"]
        var_map = cps["env_var_names"][mngr]

        # Resolve the variables from the environment, requiring them to not
        # be empty
        resolved = {
            target_key: env.get(var_name)
            for target_key, var_name in var_map.items()
            if env.get(var_name)
        }

        # Check that all required keys are available
        required = ("job_id", "num_nodes", "node_list", "node_name")
        if any([var not in resolved for var in required]):
            _missing = ", ".join([k for k in required if k not in resolved])
            raise ValueError(
                f"Missing required environment variable(s):  {_missing} ! "
                "Make sure that the corresponding environment variables are "
                "set and that the mapping is correct!\n"
                f"  Mapping for manager '{mngr}':\n{pformat(var_map)}\n\n"
                f"  Full environment:\n{pformat(env)}\n\n"
            )

        # Now do some postprocessing on some of the values
        # Ensure integers
        resolved["job_id"] = int(resolved["job_id"])
        resolved["num_nodes"] = int(resolved["num_nodes"])

        if "num_procs" in resolved:
            resolved["num_procs"] = int(resolved["num_procs"])

        if "timestamp" in resolved:
            resolved["timestamp"] = int(resolved["timestamp"])

        # Ensure reproducible node list format: ordered list
        parse_mode = self.cluster_params["node_list_parser_params"][mngr]
        try:
            node_list = parse_node_list(
                resolved["node_list"], mode=parse_mode, rcps=resolved
            )
        except Exception as exc:
            raise ValueError(
                f"Failed parsing node list {resolved['node_list']} into a "
                f"uniform format using parsing mode '{parse_mode}' and "
                f"cluster manager '{mngr}'! "
                "Check the cluster mode configuration, the relevant "
                "environment variables, and the chained error message.\n"
                f"Cluster parameters:\n{pformat(self.cluster_params)}\n\n"
                f"Parameters resolved so far:\n{pformat(resolved)}"
            ) from exc

        resolved["node_list"] = node_list

        # Calculated values, needed in Multiverse.run
        # node_index: the offset in the modulo operation
        resolved["node_index"] = node_list.index(resolved["node_name"])

        # Return the resolved values
        log.debug("Resolved cluster parameters:\n%s", pformat(resolved))
        return resolved

    def _add_sim_task(
        self, *, uni_id_str: str, uni_cfg: dict, is_sweep: bool
    ) -> None:
        """Helper function that handles task assignment to the WorkerManager.

        This function creates a WorkerTask that will perform the following
        actions **once it is grabbed and worked at**:

          - Create a universe (folder) for the task (simulation)
          - Write that universe's configuration to a yaml file in that folder
          - Create the correct arguments for the call to the model binary

        To that end, this method gathers all necessary arguments and registers
        a WorkerTask with the WorkerManager.

        Args:
            uni_id_str (str): The zero-padded uni id string
            uni_cfg (dict): given by ParamSpace. Defines how many simulations
                should be started
            is_sweep (bool): Flag is needed to distinguish between sweeps
                and single simulations. With this information, the forwarding
                of a simulation's output stream can be controlled.

        Raises:
            RuntimeError: If adding the simulation task failed
        """

        def setup_universe(
            *,
            worker_kwargs: dict,
            model_name: str,
            model_executable: str,
            uni_cfg: dict,
            uni_basename: str,
        ) -> dict:
            """The callable that will setup everything needed for a universe.

            This is called before the worker process starts working on the
            universe.

            Args:
                worker_kwargs (dict): the current status of the worker_kwargs
                    dictionary; is always passed to a task setup function
                model_name (str): The name of the model
                model_executable (str): path to the binary to execute
                uni_cfg (dict): the configuration to create a yml file from
                    which is then needed by the model
                uni_basename (str): Basename of the universe to use for folder
                    creation, i.e.: zero-padded universe number, e.g. uni0042

            Returns:
                dict: kwargs for the process to be run when task is grabbed by
                    Worker.
            """
            # Create universe directory path using the basename
            uni_dir = os.path.join(self.dirs["data"], uni_basename)

            # Now create the folder
            os.mkdir(uni_dir)
            log.debug("Created universe directory:\n  %s", uni_dir)

            # Store it in the configuration
            uni_cfg["output_dir"] = uni_dir

            # Generate a path to the output hdf5 file and add it to the dict
            output_path = os.path.join(uni_dir, "data.h5")
            uni_cfg["output_path"] = output_path

            # Parse the potentially string-valued number of steps values, and
            # other step-like arguments. Raises an error if they are negative.
            uni_cfg["num_steps"] = parse_num_steps(uni_cfg["num_steps"])
            uni_cfg["write_every"] = parse_num_steps(uni_cfg["write_every"])
            uni_cfg["write_start"] = parse_num_steps(uni_cfg["write_start"])

            # write essential part of config to file:
            uni_cfg_path = os.path.join(uni_dir, "config.yml")
            write_yml(uni_cfg, path=uni_cfg_path)

            # Build args tuple for task assignment; only need to pass the path
            # to the configuration file ...
            args = (model_executable, uni_cfg_path)

            # Generate a new worker_kwargs dict, carrying over the given ones
            wk = dict(
                args=args,
                read_stdout=True,
                stdout_parser="yaml_dict",
                **(worker_kwargs if worker_kwargs else {}),
            )

            # Determine whether to save the streams (True by default)
            if wk.get("save_streams", True):
                # Generate a path and store in the worker kwargs
                wk["save_streams_to"] = os.path.join(uni_dir, "{name:}.log")

            return wk

        # Generate the universe basename, which will be used for the folder
        # and the task name
        uni_basename = f"uni{uni_id_str}"

        # Create the dict that will be passed as arguments to setup_universe
        setup_kwargs = dict(
            model_name=self.model_name,
            model_executable=self.model_executable,
            uni_cfg=uni_cfg,
            uni_basename=uni_basename,
        )

        # Process worker_kwargs
        wk = self.meta_cfg.get("worker_kwargs")

        if wk and wk.get("forward_streams") == "in_single_run":
            # Reverse the flag to determine whether to forward streams
            wk["forward_streams"] = not is_sweep
            wk["forward_kwargs"] = dict(forward_raw=True)

        # Try to add a task to the worker manager
        try:
            self.wm.add_task(
                name=uni_basename,
                priority=None,
                setup_func=setup_universe,
                setup_kwargs=setup_kwargs,
                worker_kwargs=wk,
            )

        except RuntimeError as err:
            # Task list was locked, probably because there already was a run
            raise RuntimeError(
                "Could not add simulation task for universe '{uni_basename}'! "
                "Did you already perform a run with this Multiverse?"
            ) from err

        log.debug("Added simulation task: %s.", uni_basename)

    def _add_sim_tasks(self, *, sweep: bool = None) -> int:
        """Adds the simulation tasks needed for a single run or for a sweep.

        Args:
            sweep (bool, optional): Whether tasks for a parameter sweep should
                be added or only for a single universe. If None, will read the
                ``perform_sweep`` key from the meta-configuration.

        Returns:
            int: The number of added tasks.

        Raises:
            ValueError: On ``sweep == True`` and zero-volume parameter space.
        """
        if sweep is None:
            sweep = self.meta_cfg.get("perform_sweep", False)

        pspace = self.meta_cfg["parameter_space"]

        if not sweep:
            # Only need the default state of the parameter space
            uni_cfg = pspace.default

            # Make a backup of the parameter space that is *actually* used
            self._perform_pspace_backup(
                psp.ParamSpace(uni_cfg),
                filename="parameter_space",
                perform_sweep=False,
            )

            # Add the task to the worker manager.
            log.progress("Adding task for simulation of a single universe ...")
            self._add_sim_task(uni_id_str="0", uni_cfg=uni_cfg, is_sweep=False)

            return 1
        # -- else: tasks for parameter sweep needed

        if pspace.volume < 1:
            raise ValueError(
                "The parameter space has no sweeps configured! "
                "Refusing to run a sweep. You can either call "
                "the run_single method or add sweeps to your "
                "run configuration using the !sweep YAML tags."
            )

        # Get the parameter space iterator and the number of already-existing
        # tasks (to later compute the number of _added_ tasks)
        psp_iter = pspace.iterator(with_info="state_no_str")
        _num_tasks = len(self.wm.tasks)

        # Distinguish whether to do a regular sweep or we are in cluster mode
        if not self.cluster_mode:
            # Make a backup of the parameter space that is *actually* used
            self._perform_pspace_backup(
                pspace, filename="parameter_space", perform_sweep=True
            )

            # Do a sweep over the whole activated parameter space
            vol = pspace.volume
            log.progress(
                "Adding tasks for simulation of %d universes ...", vol
            )

            for i, (uni_cfg, uni_id_str) in enumerate(psp_iter):
                self._add_sim_task(
                    uni_id_str=uni_id_str, uni_cfg=uni_cfg, is_sweep=True
                )

                print(
                    f"  Added simulation task:  {uni_id_str}  ({i+1}/{vol})",
                    end="\r",
                )

        else:
            # Prepare a cluster mode sweep
            log.info("Preparing cluster mode sweep ...")

            # Get the resolved cluster parameters
            # These include the following values:
            #    num_nodes:   The total number of nodes to simulate on. This
            #                 is what determines the modulo value.
            #    node_index:  Equivalent to the modulo offset, which depends
            #                 on the position of this Multiverse's node in the
            #                 sequence of all nodes.
            rcps = self.resolved_cluster_params
            num_nodes = rcps["num_nodes"]
            node_index = rcps["node_index"]

            # Back up the actually-used parameter space. Do this only on the
            # first node to avoid file-writing conflicts between nodes
            if node_index == 0:
                self._perform_pspace_backup(
                    pspace, filename="parameter_space", perform_sweep=True
                )

            # Inform about the number of universes to be simulated
            log.progress(
                "Adding tasks for cluster-mode simulation of "
                "%d universes on this node (%d of %d) ...",
                (
                    pspace.volume // num_nodes
                    + (pspace.volume % num_nodes > node_index)
                ),
                node_index + 1,
                num_nodes,
            )

            for i, (uni_cfg, uni_id_str) in enumerate(psp_iter):
                # Skip if this node is not responsible
                if (i - node_index) % num_nodes != 0:
                    log.debug("Skipping:  %s", uni_id_str)
                    continue

                # Is valid for this node, add the simulation task
                self._add_sim_task(
                    uni_id_str=uni_id_str, uni_cfg=uni_cfg, is_sweep=True
                )

        num_new_tasks = len(self.wm.tasks) - _num_tasks
        log.info("Added %d tasks.", num_new_tasks)
        return num_new_tasks

    def _validate_meta_cfg(self) -> bool:
        """Goes through the parameters that require validation, validates them,
        and creates a useful error message if there were invalid parameters.

        Returns:
            bool: True if all parameters are valid; None if no check was done.
                Note that False will never be returned, but a ValidationError
                will be raised instead.

        Raises:
            ValidationError: If validation failed.
        """
        to_validate = self.meta_cfg.get("parameters_to_validate", {})

        if not to_validate:
            log.info("Skipping parameter validation: nothing to validate.")
            return None

        elif not self.meta_cfg.get("perform_validation", True):
            log.info("Skipping parameter validation: is deactivated.")
            return None

        log.info("Validating %d parameters ...", len(to_validate))
        pspace = self.meta_cfg["parameter_space"]
        log.remark("Parameter space volume:      %d", pspace.volume)

        if pspace.volume >= 1000:
            log.note(
                "This may take a few seconds. To skip validation, set "
                "the `perform_validation` flag to False."
            )

        # The dict to collect details on invalid parameters in.
        #   - Keys are key sequences (tuple of str)
        #   - Values are sets of error _messages_, hence suppressing duplicates
        invalid_params = defaultdict(set)

        # Iterate over the whole parameter space, including the default point
        # TODO Can improve performance by directly checking sweep dimensions.
        #      ... but be very careful about robustness here!
        for params in itertools.chain(pspace, [pspace.default]):
            for key_seq, param in to_validate.items():
                # Retrieve the value from this point in parameter space
                value = psp.tools.recursive_getitem(params, keys=key_seq)

                # Validate it and store the error _message_ if invalid
                try:
                    param.validate(value)

                except ValidationError as exc:
                    invalid_params[key_seq].add(str(exc))

        if not invalid_params:
            log.note("All parameters valid.")
            return True

        # else: Validation failed. Create an informative error message
        msg = (
            f"Validation failed for {len(invalid_params)} "
            f"parameter{'s' if len(invalid_params) > 1 else ''}:\n\n"
        )

        # Get the length of longest key sequence (used for alignment)
        _nd = max(len(".".join(ks)) for ks in invalid_params.keys())

        for key_seq, errs in invalid_params.items():
            path = ".".join(key_seq)

            if len(errs) == 1:
                msg += f"  - {path:<{_nd}s}  :  {list(errs)[0]}\n"
            else:
                _details = "\n".join([f"     - {e}" for e in errs])
                msg += (
                    f"  - {path:<{_nd}s}  :  validation failed for "
                    f"{len(errs)} sweep values:\n{_details}\n"
                )

        msg += (
            "\nInspect the details above and adjust the run configuration "
            "accordingly.\n"
        )
        raise ValidationError(msg)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class FrozenMultiverse(Multiverse):
    """A frozen Multiverse is like a Multiverse, but frozen.

    It is initialized from a finished :py:class:`~utopya.multiverse.Multiverse`
    run and re-creates all the attributes from that data, e.g.: the meta
    configuration, a DataManager, and a PlotManager.

    .. note::

        A frozen multiverse is no longer able to perform any simulations.
    """

    def __init__(
        self,
        *,
        model_name: str = None,
        info_bundle: ModelInfoBundle = None,
        run_dir: str = None,
        run_cfg_path: str = None,
        user_cfg_path: str = None,
        use_meta_cfg_from_run_dir: bool = False,
        **update_meta_cfg,
    ):
        """Initializes the FrozenMultiverse from a model name and the name
        of a run directory.

        Note that this also takes arguments to specify the run configuration to
        use.

        Args:
            model_name (str): The name of the model to load. From this, the
                model output directory is determined and the run_dir will be
                seen as relative to that directory.
            info_bundle (ModelInfoBundle, optional): The model information
                bundle that includes information about the binary path etc.
                If not given, will attempt to read it from the model registry.
            run_dir (str, optional): The run directory to load. Can be a path
                relative to the current working directory, an absolute path,
                or the timestamp of the run directory. If not given, will use
                the most recent timestamp.
            run_cfg_path (str, optional): The path to the run configuration.
            user_cfg_path (str, optional): If given, this is used to update the
                base configuration. If None, will look for it in the default
                path, see Multiverse.USER_CFG_SEARCH_PATH.
            use_meta_cfg_from_run_dir (bool, optional): If True, will load the
                meta configuration from the given run directory; only works for
                absolute run directories.
            **update_meta_cfg: Can be used to update the meta configuration
                generated from the previous configuration levels
        """
        # First things first: get the info bundle
        self._info_bundle = get_info_bundle(
            model_name=model_name, info_bundle=info_bundle
        )
        log.progress(
            "Initializing FrozenMultiverse for '%s' model ...", self.model_name
        )

        # Initialize property-managed attributes
        self._meta_cfg = None
        self._dirs = dict()
        self._resolved_cluster_params = None

        # Decide whether to load the meta configuration from the given run
        # directory or the currently available one.
        if (
            use_meta_cfg_from_run_dir
            and isinstance(run_dir, str)
            and os.path.isabs(run_dir)
        ):
            raise NotImplementedError("use_meta_cfg_from_run_dir")

            # Find the meta config backup file and load it
            # Alternatively, create it from the singular backup files ...
            # log.info("Trying to load meta configuration from given absolute "
            #          "run directory ...")
            # Update it with the given update_meta_cfg dict

        else:
            # Need to create a meta configuration from the currently available
            # values.
            mcfg, _ = self._create_meta_cfg(
                run_cfg_path=run_cfg_path,
                user_cfg_path=user_cfg_path,
                update_meta_cfg=update_meta_cfg,
            )

        # Only keep selected entries from the meta configuration. The rest is
        # not needed and is deleted in order to not confuse the user with
        # potentially varying versions of the meta config.
        self._meta_cfg = {
            k: v
            for k, v in mcfg.items()
            if k
            in (
                "debug_level",
                "paths",
                "data_manager",
                "plot_manager",
                "cluster_mode",
                "cluster_params",
            )
        }
        log.info("Built meta configuration.")
        log.remark("  Debug level:  %d", self.debug_level)
        self._apply_debug_level()

        # Need to make some DataManager adjustments; do so via update dicts
        dm_cluster_kwargs = dict()
        if self.cluster_mode:
            log.note("Cluster mode enabled.")
            self._resolved_cluster_params = self._resolve_cluster_params()
            rcps = self.resolved_cluster_params  # creates a deep copy

            log.note(
                "This is node %d of %d.",
                rcps["node_index"] + 1,
                rcps["num_nodes"],
            )

            # Changes to the meta configuration
            # To avoid config file collisions in the PlotManager:
            self._meta_cfg["plot_manager"]["cfg_exists_action"] = "skip"

            # _Additional_ arguments to pass to DataManager.__init__ below
            timestamp = rcps["timestamp"]
            dm_cluster_kwargs = dict(
                out_dir_kwargs=dict(timestamp=timestamp, exist_ok=True)
            )

        # Generate the path to the run directory that is to be loaded
        self._create_run_dir(**self.meta_cfg["paths"], run_dir=run_dir)
        log.note("Run directory:\n  %s", self.dirs["run"])

        # Create a data manager
        self._dm = DataManager(
            self.dirs["run"],
            name=f"{self.model_name}_data",
            **self.meta_cfg["data_manager"],
            **dm_cluster_kwargs,
        )
        log.progress("Initialized DataManager.")

        # Instantiate the PlotManager via the helper method
        self._pm = self._setup_pm()

        log.progress("Initialized FrozenMultiverse.\n")

    def _create_run_dir(self, *, out_dir: str, run_dir: str, **__):
        """Helper function to find the run directory from arguments given
        to :py:meth:`~utopya.multiverse.Multiverse.__init__`.

        Overwrites the method from the parent Multiverse class, because the
        FrozenMultiverse does not require setting up a *new* run directory but
        should instead identify the existing one and create an appropriate
        output directory.

        Args:
            out_dir (str): The output directory
            run_dir (str): The run directory to use
            ``**__``: ignored

        Raises:
            IOError: No directory found to use as run directory
            TypeError: When run_dir was not a string
        """
        PATTERN = r"\d{6}-\d{6}_?.*"

        # Create model directory path (where the to-be-loaded data is expected)
        out_dir = os.path.expanduser(str(out_dir))
        model_dir = os.path.join(out_dir, self.model_name)

        if not os.path.isdir(model_dir):
            # Just create it, there's no harm in that ...
            os.makedirs(model_dir)

        # Distinguish different types of values for the run_dir argument
        if run_dir is None:
            log.info("Trying to identify the most recent run directory ...")

            # Create list of _directories_ matching timestamp pattern
            dirs = [
                d
                for d in sorted(os.listdir(model_dir))
                if os.path.isdir(os.path.join(model_dir, d))
                and re.match(PATTERN, os.path.basename(d))
            ]

            if not dirs:
                raise FileNotFoundError(
                    "Could not find a run directory to load for evaluation "
                    f"of model '{self.model_name}'!\n"
                    f"Model output directory:  {model_dir}\n\n"
                    "Did you perform a simulation yet? If you are using this "
                    "model only for evaluation, place your data in a new "
                    "subdirectory in that folder (using timestamp as name)."
                )

            # Use the latest to choose the run directory
            run_dir = os.path.join(model_dir, dirs[-1])

        elif isinstance(run_dir, str):
            run_dir = os.path.expanduser(run_dir)

            # Distinguish absolute and relative paths and those starting with
            # a timestamp-like pattern, which can be looked up from the model
            # directory.
            if os.path.isabs(run_dir):
                log.debug("Received absolute run_dir, using that one.")

            elif re.match(PATTERN, run_dir):
                # Looks like a relative path within the model directory
                log.info(
                    "Received timestamp '%s' for run_dir; trying to find "
                    "one within the model output directory ...",
                    run_dir,
                )
                run_dir = os.path.join(model_dir, run_dir)

            else:
                # Is not an absolute path and not a timestamp; thus a path
                # relative to the current working directory
                run_dir = os.path.join(os.getcwd(), run_dir)

        else:
            raise TypeError(
                "Argument run_dir needs to be None, an absolute "
                "path, or a path relative to the model output "
                f"directory, but it was: {run_dir}"
            )

        # Check if the directory exists
        if not os.path.isdir(run_dir):
            raise OSError(f"No run directory found at '{run_dir}'!")

        # Store the path and associate the subdirectories
        self.dirs["run"] = run_dir

        for subdir in ("config", "eval", "data"):
            subdir_path = os.path.join(run_dir, subdir)
            self.dirs[subdir] = subdir_path
            # TODO Consider checking if it exists?



"""Implements signalling-related functionality and globally relevant data"""

import signal as _signal

SIGMAP = {a: int(getattr(_signal, a)) for a in dir(_signal) if a[:3] == "SIG"}
"""A map from signal names to corresponding integer exit codes"""



"""Tools for model registration"""

import logging

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


def register_models_from_list(
    *,
    registry: "utopya.model_registry.registry.ModelRegistry",
    separator: str,
    model_names: str,
    executables: str,
    label: str,
    more_paths: dict = dict(),
    source_dirs: str = None,
    exists_action: str = "raise",
    set_as_default: bool = None,
    project_name: str = None,
    _log=log,
    **shared_bundle_kwargs,
):
    """Handles registration of multiple models where the model names,
    executables, and source directories are splittable lists of equal lengths.

    Args:
        registry (utopya.model_registry.registry.ModelRegistry): The model
            registry to store the models in
        separator (str): Separation string to split ``model_names``,
            ``executables``, and ``source_dirs``.
        model_names (str): Splittable string of model names
        executables (str): Splittable string of executables
        label (str): Label under which to add the entries
        more_paths (dict, optional): Additional paths that are to be parsed
        source_dirs (str, optional): Splittable string of model source
            directories
        exists_action (str, optional): Action to take upon existing label
        project_name (str, optional): The associated project name
        _log (logging.Logger, optional): A logger-like object
        **shared_bundle_kwargs: passed on to bundle creation
    """

    _log.debug(
        "Splitting given model registration arguments by '%s' ...",
        separator,
    )

    model_names = model_names.split(separator)
    executables = executables.split(separator)

    if not source_dirs:
        source_dirs = [None for _ in model_names]
    else:
        source_dirs = source_dirs.split(separator)

    if not (len(model_names) == len(executables) == len(source_dirs)):
        raise ValueError(
            "Mismatch of sequence lengths during list-based model "
            "registration! The model_names, executables, and source_dirs "
            "lists should all be of equal length after having been split "
            f"by separator '{separator}', but were:\n"
            f"  model_names ({len(model_names)}):  {model_names}\n"
            f"  executables ({len(executables)}):  {executables}\n"
            f"  source_dirs ({len(source_dirs)}):  {source_dirs}"
        )

    # Go over them, create the paths dict, and populate a specs dict.
    specs = dict()
    for model_name, executable, source_dir in zip(
        model_names, executables, source_dirs
    ):
        paths = dict(
            executable=executable,
            source_dir=source_dir,
            **more_paths,
        )

        specs[model_name] = dict(
            paths=paths, project_name=project_name, **shared_bundle_kwargs
        )

    _log.progress(
        "Received information for %d model%s. Now registering ...\n",
        len(specs),
        "s" if len(specs) != 1 else "",
    )

    # Now register, passing along shared arguments
    for model_name, bundle_kwargs in specs.items():
        registry.register_model_info(
            model_name,
            **bundle_kwargs,
            label=label,
            set_as_default=set_as_default,
            exists_action=exists_action,
        )

    _log.success("\nModel registration succeeded.")
    _log.remark(registry.info_str)



"""This module provides matplotlib-related helper constructs"""

import copy
import logging
from typing import Tuple, Union

import matplotlib as mpl
import matplotlib.patches as mpatches
import numpy as np
from dantro.plot import ColorManager
from matplotlib.legend_handler import HandlerPatch

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


def adjust_figsize_to_aspect(
    aspect: float, *, fig: "matplotlib.figure.Figure"
) -> Tuple[float, float]:
    """Adjusts the given figures size, *enlarging* it to match the given aspect
    ratio where ``width = height * aspect``.

    Args:
        aspect (float): The aspect ratio ``width / height``
        fig (matplotlib.figure.Figure): The figure

    Returns:
        Tuple[float, float]:
            New width and height of the figure in inches.
    """
    w, h = fig.get_size_inches()
    if w < h * aspect:
        w = h * aspect
    else:
        h = w / aspect
    fig.set_size_inches(w, h)
    return fig.get_size_inches()


class HandlerEllipse(HandlerPatch):
    """Custom legend handler to turn an ellipse handle into a legend key."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create_artists(
        self,
        legend,
        orig_handle,
        xdescent,
        ydescent,
        width,
        height,
        fontsize,
        trans,
    ):
        """Create an ellipse as a matplotlib artist object."""
        center = 0.5 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent
        p = mpatches.Ellipse(
            xy=center, width=height + xdescent, height=height + ydescent
        )
        self.update_prop(p, orig_handle, legend)
        p.set_transform(trans)
        return [p]



"""This module provides plotting functions to visualize cellular automata."""

import copy
import logging
import warnings
from math import ceil, sqrt
from typing import Callable, Dict, Optional, Sequence, Tuple, Union

import matplotlib as mpl
import matplotlib.collections
import matplotlib.image
import matplotlib.patches
import matplotlib.transforms
import numpy as np
import xarray as xr
from dantro.abc import AbstractDataContainer
from dantro.plot import ColorManager
from dantro.plot.funcs.generic import make_facet_grid_plot
from matplotlib.colors import ListedColormap

from ...tools import ensure_dict, recursive_update
from .. import DataManager, UniverseGroup
from . import PlotHelper, UniversePlotCreator, is_plot_func

log = logging.getLogger(__name__)

# Increase log threshold for animation module
logging.getLogger("matplotlib.animation").setLevel(logging.WARNING)


# -----------------------------------------------------------------------------


def _prepare_hexgrid_data(
    data: Union[np.ndarray, xr.DataArray], *, x: str = None, y: str = None
) -> Tuple[xr.DataArray, str, str]:
    """Prepares data for :py:func:`imshow_hexagonal` by checking the given data
    and specified dimension names are consistent.
    """
    if not isinstance(data, xr.DataArray):
        data = xr.DataArray(data)

    if data.ndim != 2:
        raise ValueError(
            "Need 2-dimensional data for hexagonal grid plot, but got "
            f"{data.ndim}-dimensional data!\n{str(data)}"
        )

    # May not have dimension arguments given, e.g. because this was a data
    # array. Assume that it's fine to just use the available two.
    # For simplicity, we don't allow to give only one of them.
    if not x and not y:
        x, y = data.dims

    elif x == y:
        raise ValueError(
            "Dimension names `x` and `y` need to be different, "
            f"but are both '{x}'!"
        )

    elif (x is None) != (y is None):
        raise ValueError(
            "Need either both `x` and `y` dimension names or neither!"
        )

    # Make sure data dimensions are ordered correctly
    data = data.transpose(x, y)

    return data, x, y


def _flatten_hexgrid_data(data: xr.DataArray) -> np.ndarray:
    """Flattens hexgrid data in a specific way.

    For consistency, this function should be used when calling
    :py:meth:`~matplotlib.collections.PolyCollection.set_array` on collections
    that represent hexagonal grid data.
    """
    return data.data.T.flatten()


def _plot_ca_property(
    prop_name: str,
    *,
    hlpr: PlotHelper,
    data: xr.DataArray,
    default_imshow_kwargs: dict,
    imshow_hexagonal_extra_kwargs: dict = None,
    default_cbar_kwargs: dict = None,
    grid_structure: str = None,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    cmap: Union[str, dict] = None,
    norm: Union[str, dict] = None,
    add_colorbar: bool = True,
    set_axis_off: bool = True,
    title: str = None,
    imshow_kwargs: dict = None,
    cbar_labels: dict = None,
    cbar_label_kwargs: dict = None,
    cbar_tick_params: dict = None,
    no_cbar_markings: bool = False,
    **cbar_kwargs,
) -> mpl.image.AxesImage:
    """Helper function, used in :py:func:`caplot` to plot a property on the
    given axis. Returns the created axes image object.

    .. note::

        The arguments here are those within the individual entries of the
        ``to_plot`` argument for the above plotting functions.

    Args:
        prop_name (str): The property to plot
        hlpr (PlotHelper): The plot helper
        data (xarray.DataArray): The array-like data to plot as an image
        default_imshow_kwargs (dict): Default arguments for the imshow call,
            updated with individually-specified ``imshow_kwargs``.
        imshow_hexagonal_extra_kwargs (dict): Default arguments for hexagonal
            grids, ignored otherwise.
            This updates the ``default_imshow_kwargs`` and is in turn updated
            with individually-specified ``imshow_kwargs``.
        default_cbar_kwargs (dict): Default arguments for the colorbar
            creation, updated with ``cbar_kwargs``.
        grid_structure (str, optional): Can be used to explicitly set the grid
            structure in cases where ``data.attrs['grid_structure']`` is not
            available or holds an invalid entry.
            This decides whether to use :py:meth:`matplotlib.axes.Axes.imshow`
            or :py:func:`imshow_hexagonal`.
            Note that the ``grid_properties`` need to be passed via the
            ``imshow_kwargs`` argument below.
        vmin (float, optional): The lower limit to use for the colorbar range.
        vmax (float, optional): The upper limit to use for the colorbar range.
        cmap (Union[str, dict], optional): The colormap to use. If a dict is
            given, defines a (discrete) ``ListedColormap`` from the values.
            Handled by :py:class:`~dantro.plot.utils.color_mngr.ColorManager`.
        norm (Union[str, dict], optional): The normalization function to use.
            Handled by :py:class:`~dantro.plot.utils.color_mngr.ColorManager`.
        add_colorbar (bool, optional): If false, will not draw a colorbar.
            Default is true.
        set_axis_off (bool, optional): If true (default), will set the axis to
            invisible.
        title (str, optional): The subplot figure title
        imshow_kwargs (dict, optional): Depending on grid structure, is passed
            on either to :py:meth:`~matplotlib.axes.Axes.imshow` or to
            :py:func:`.imshow_hexagonal`.
        cbar_labels (dict, optional): Passed to
            :py:class:`~dantro.plot.utils.color_mngr.ColorManager` to set up
            the label names alongside the given ``cmap`` and ``norm``.
        cbar_label_kwargs (dict, optional): Passed to
            :py:meth:`~dantro.plot.utils.color_mngr.ColorManager.create_cbar`
            for controlling the aesthetics of colorbar labels.
        cbar_tick_params (dict, optional): Passed to
            :py:meth:`~dantro.plot.utils.color_mngr.ColorManager.create_cbar`
            for controlling the aesthetics of colorbar ticks.
        no_cbar_markings (bool, optional): Whether to suppress colorbar
            markings (ticks and tick labels).
        **cbar_kwargs: Passed to
            :py:meth:`~dantro.plot.utils.color_mngr.ColorManager.create_cbar`

    Returns:
        matplotlib.image.AxesImage:
            The created axes image representing the CA property.

    Raises:
        ValueError: on invalid grid structure; supported structures are
            ``square`` and ``hexagonal``
    """
    # Set up the ColorManager
    cm = ColorManager(
        cmap=cmap, norm=norm, vmin=vmin, vmax=vmax, labels=cbar_labels
    )

    # Determine grid structure
    grid_structure = (
        grid_structure
        if grid_structure
        else data.attrs.get("grid_structure", "square")
    )

    # Prepare imshow_kwargs, successively updating defaults.
    # Also need to be able to pass custom arguments to imshow_hexagonal, which
    # has a wider interface than regular imshow ...
    _imshow_kwargs = ensure_dict(default_imshow_kwargs)
    if grid_structure == "hexagonal":
        _imshow_kwargs = recursive_update(
            _imshow_kwargs, ensure_dict(imshow_hexagonal_extra_kwargs)
        )
    _imshow_kwargs = recursive_update(
        _imshow_kwargs, ensure_dict(imshow_kwargs)
    )

    # Create imshow(-like) object on the currently selected axis
    if grid_structure == "square" or grid_structure is None:
        im = hlpr.ax.imshow(
            data.T,
            cmap=cm.cmap,
            norm=cm.norm,
            animated=True,
            rasterized=True,
            origin="lower",
            aspect="equal",
            **_imshow_kwargs,
        )

    elif grid_structure == "hexagonal":
        im = imshow_hexagonal(
            data=data,
            ax=hlpr.ax,
            cmap=cm.cmap,
            norm=cm.norm,
            animated=True,
            rasterized=True,
            **_imshow_kwargs,
        )

    else:
        raise ValueError(
            f"Unsupported grid structure '{grid_structure}'!\n"
            "Choose from:  square, hexagonal"
        )

    # Remove main axis labels and ticks and provide some default options
    if set_axis_off:
        hlpr.ax.axis("off")

    hlpr.provide_defaults("set_title", title=(title if title else prop_name))

    # .. Colorbar .............................................................
    if not add_colorbar:
        return im
    # else: draw the colorbar

    # Determine which artist to use; for hexagonal grids, need to attach the
    # PolyCollection, because it holds the data array.
    artist = im
    if grid_structure == "hexagonal":
        artist = im.hexagons

    # Parse colorbar kwargs, setting some default values
    default_cbar_kwargs = ensure_dict(default_cbar_kwargs)
    cbar_kwargs = recursive_update(
        copy.deepcopy(default_cbar_kwargs), cbar_kwargs
    )
    cbar_kwargs["fraction"] = cbar_kwargs.get("fraction", 0.05)
    cbar_kwargs["pad"] = cbar_kwargs.get("pad", 0.02)

    # Draw the colorbar, then store it in the AxesImage to have it accesible
    cbar = cm.create_cbar(
        artist,
        fig=hlpr.fig,
        ax=hlpr.ax,
        label_kwargs=cbar_label_kwargs,
        tick_params=cbar_tick_params,
        **cbar_kwargs,
    )
    im.cbar = cbar

    # May want to remove markings
    if no_cbar_markings:
        cbar.set_ticks([])
        cbar.ax.set_xticklabels([])
        cbar.ax.set_yticklabels([])

    return im


# -----------------------------------------------------------------------------
# -- Plot functions -----------------------------------------------------------
# -----------------------------------------------------------------------------


def imshow_hexagonal(
    data: Union[xr.DataArray, np.ndarray],
    *,
    ax: "matplotlib.axes.Axes" = None,
    x: str = None,
    y: str = None,
    grid_properties: dict = {},
    update_grid_properties: dict = {},
    grid_properties_keys: dict = {},
    extent: tuple = None,
    scale: float = 1.01,
    draw_centers: bool = False,
    draw_center_radius: float = 0.1,
    draw_center_kwargs: dict = {},
    hide_ticks: bool = None,
    cmap: str = None,
    norm: str = None,
    vmin: float = None,
    vmax: float = None,
    collection_kwargs: dict = {},
    **im_kwargs,
) -> mpl.image.AxesImage:
    """Visualizes data using a grid of hexagons (⬢ or ⬣).

    Owing to the many ways in which a hexagonal grid can be visualized,
    this function requires more information than
    :py:meth:`~matplotlib.axes.Axes.imshow`. These so called *grid properties*
    need to be passed via the ``grid_properties`` argument or directly
    alongside the data via ``data.attrs`` (for :py:class:`~xarray.DataArray`).

    The following grid properties are available:

        coordinate_mode (str):
            In which way the data is stored. Currently only supports
            ``offset`` mode, i.e. with offset row and column coordinates.
            Coordinates of individual cells need not be given, nor can they
            be given: Assuming a regular hexagonal grid, all coordinates and
            sizes are completely deduced from the shape of the given data and
            the other grid parameters like ``pointy_top`` and ``offset_mode``.
        pointy_top (bool):
            Whether the hexagons have a pointy top (⬢) or a flat top (⬣).
            More precisely, with a pointy top, there is only a single vertex
            at the top and bottom of the hexagon (i.e. along the ``y``
            dimension and the top/bottom of the resulting plot).
        offset_mode (str):
            Whether ``even`` or ``odd`` rows or columns are offset towards
            higher values. In other words:
            For pointy tops, offset every second *row* toward the right.
            For flat tops, offset every second *column* towards the top.
            Offset distance is half a cell's width and half a cell's height,
            respectively.
        space_size (Tuple[float, float], optional):
            The size of the space in ``(width, height)`` that the hexagonal
            grid cells cover.
            If given, will make the assumption that the available number of
            hexagons in each dimension reach from one end of the space to the
            other, even if that means that the hexagons become distorted along
            the dimensions.
            If *not* given, will assume regular hexagons and use arbitrary
            units to cover the space; these hexagons will *not* be distorted.
        space_offset (Tuple[float, float], optional):
            Translates the space by ``(offset_x, offset_y)``.
            Also applies to the case where ``space_size`` was not given.
            Effectively, this refers to the coordinates of the bottom left-hand
            corner of the space.
        space_boundary (str, optional):
            Whether the space (regardless of explicitly given or deduced)
            describes the ``outer`` or ``inner`` boundary of the hexagonal
            grid. The ``outer`` boundary (default) goes through the outermost
            vertices of the outermost cells.
            The ``inner`` boundary goes through some hexagon center, cutting
            off pointy tops and protruding parts of the hexagons such that the
            whole space is covered by hexagonal cells.

    For example, grid properties may look like this:

    .. code-block:: yaml

        grid_properties:
          # -- Required:
          coordinate_mode: offset
          pointy_top: true
          offset_mode: even

          # -- Optional:
          space_size: [8, 8]
          space_offset: [-4, -4]
          space_boundary: outer

    With some 2D dummy grid data of shape ``(21, 24)``, the corresponding
    output would be as follows.
    The darker cells denote the boundary and the corners; the lighter cells
    correspond to a "vertical" line in the third column of the grid.

    .. image:: ../_static/_gen/imshow_hex/small_with_space_outer.pdf
        :target: ../_static/_gen/imshow_hex/small_with_space_outer.pdf
        :width: 100%

    By setting the space boundary parameter to ``inner``, the domain size
    remains the same, but the boundary cells are partially cut off:

    .. image:: ../_static/_gen/imshow_hex/small_with_space_inner.pdf
        :target: ../_static/_gen/imshow_hex/small_with_space_inner.pdf
        :width: 100%

    Removing all optional parameters, specifically the ``space_size``, the
    size of the domain is arbitrary, so no labels are drawn.
    In addition, we can also mark the hexagon centers:

    .. image:: ../_static/_gen/imshow_hex/small_centers_marked.pdf
        :target: ../_static/_gen/imshow_hex/small_centers_marked.pdf
        :width: 100%

    Changing to flat tops and ``odd`` offset mode results in a figure with a
    different aspect ratio, while the hexagons remain regular.

    .. image:: ../_static/_gen/imshow_hex/small_flat_top_odd.pdf
        :target: ../_static/_gen/imshow_hex/small_flat_top_odd.pdf
        :width: 100%

    When specifying the domain size again, the hexagons need to be scaled
    non-uniformly to cover the domain:

    .. image:: ../_static/_gen/imshow_hex/small_flat_top_odd_with_space.pdf
        :target: ../_static/_gen/imshow_hex/small_flat_top_odd_with_space.pdf
        :width: 100%

    .. hint::

        For an excellent introduction to hexagonal grid representations, see
        `this article <https://www.redblobgames.com/grids/hexagons/>`_.

    .. admonition:: See also

        * :py:func:`caplot` integrates this function.
        * :ref:`plot_funcs_ca_hex` documents usage and shows more examples.

    Args:
        data (Union[xarray.DataArray, numpy.ndarray]): 2D array-like data that
            holds the grid information that is to be plotted.
            If the data is given as :py:class:`~xarray.DataArray`, its
            ``attrs`` are used to *update* the given ``grid_properties``.
        ax (matplotlib.axes.Axes, optional): The axes to draw to; if not given,
            will use the current axes.
        x (str, optional): Name of the data dimension that is to be represented
            on the x-axis of the plot. If not given, will use the first data
            dimension.
        y (str, optional): Name of the data dimension that is to be represented
            on the x-axis of the plot. If not given, will use the second data
            dimension.
        grid_properties (dict, optional): The grid properties dict, which needs
            to specify the above properties in order to determine how to
            represent the data. This dict is first updated with potentially
            available ``data.attrs`` and subsequently updated with the
            ``update_grid_properties``.
        update_grid_properties (dict, optional): Updates the grid properties,
            see above.
        grid_properties_keys (dict, optional): A mapping that can be used if
            the given data has grid properties given under different names.
            For instance, ``{"space_size": "size"}`` would read the ``size``
            entry instead of ``space_size``.
        extent (tuple, optional): A custom space *extent*, denoting the edges
            ``(left, right, bottom, top)`` of the domain *in data units*.
        scale (float, optional): A scaling factor for the size of the hexagons.
            The default value is very slightly larger than 1 to reduce aliasing
            artefacts on exactly overlapping hexagon edges. Scaling is uniform.
        draw_centers (bool, optional): Whether to additionally draw the center
            points of all hexagons.
        draw_center_radius (float, optional): The relative radius of the
            center points in units of ``min(cell_width, cell_height) / 2``.
        draw_center_kwargs (dict, optional): Additional arguments that are
            passed to the :py:class:`~matplotlib.collections.PatchCollection`
            used to draw the center points.
        hide_ticks (bool, optional): Whether to hide the ticks and tick labels.
            If None, will hide ticks if no ``space_size`` grid property was
            given (in which case the units are assumed irrelevant).
        cmap (str, optional): The colormap to use
        norm (str, optional): The normalization to use
        vmin (float, optional): The minimum value of the color range to use
        vmax (float, optional): The maximum value of the color range to use
        collection_kwargs (dict, optional): Passed on to the
            :py:class:`~matplotlib.collections.PolyCollection` that is used to
            represent the hexagons.
        **im_kwargs: Passed on to :py:class:`matplotlib.image.AxesImage` that
            is created from the whole axis.
            Can be used to set ``interpolation`` or similar options.

    Returns:
        matplotlib.image.AxesImage:
            The imshow-like object containing the hexagonal grid.
    """
    # .. constants ............................................................
    sqrt_3 = np.sqrt(3)  # 2 * sin(60°)

    # Regular unit hexagon vertices in clockwise direction.
    # Define one with a pointy and one with a flat top (easier than rotating).
    unit_hexagon_pointy_top = np.array(
        [
            [0, 1],
            [sqrt_3 / 2, +1 / 2],
            [sqrt_3 / 2, -1 / 2],
            [0, -1],
            [-sqrt_3 / 2, -1 / 2],
            [-sqrt_3 / 2, +1 / 2],
        ]
    )
    unit_hexagon_flat_top = np.array(
        [
            [1, 0],
            [+1 / 2, sqrt_3 / 2],
            [-1 / 2, sqrt_3 / 2],
            [-1, 0],
            [-1 / 2, -sqrt_3 / 2],
            [+1 / 2, -sqrt_3 / 2],
        ]
    )

    # .. Prepare data .........................................................

    # Bring data into a uniform shape: 2D xr.DataArray
    data, x, y = _prepare_hexgrid_data(data, x=x, y=y)

    # Aggregate grid properties
    grid_properties = ensure_dict(copy.deepcopy(grid_properties))
    grid_properties.update(data.attrs)
    grid_properties.update(ensure_dict(copy.deepcopy(update_grid_properties)))
    if not grid_properties:
        raise ValueError(
            "Could not determine grid properties! "
            "Either pass them explicitly via the `grid_properties` or "
            "`update_grid_properties` arguments or, if `data` is given as "
            "an xr.DataArray, add them to `data.attrs`."
        )

    # .. Get hexgrid information ..............................................
    GRID_PROP_KEYS = (
        "coordinate_mode",
        "pointy_top",
        "offset_mode",
        "space_size",
        "space_offset",
        "space_boundary",
    )
    _keys = {k: grid_properties_keys.get(k, k) for k in GRID_PROP_KEYS}

    # Extract attribute values and give a useful error message if that fails
    try:
        coordinate_mode = grid_properties[_keys["coordinate_mode"]]
        pointy_top = grid_properties[_keys["pointy_top"]]
        offset_mode = grid_properties[_keys["offset_mode"]]
        space_size = grid_properties.get(_keys["space_size"])
        space_given = space_size is not None
        space_offset = grid_properties.get(_keys["space_offset"], (0.0, 0.0))
        boundary = grid_properties.get(_keys["space_boundary"], "outer")

    except KeyError as err:
        _gp = "\n".join(f"  {k:18}: {v}" for k, v in grid_properties.items())
        _km = "\n".join(f"  {k:16} -> {v}" for k, v in _keys.items())
        raise ValueError(
            f"Missing grid property {err} for imshow_hexagonal! Make sure the "
            "required metadata is available and the key mapping is correct.\n"
            f"Data attributes:\n{data.attrs}"
            f"\n\nAggregated grid properties (after updates):\n{_gp}"
            f"\n\nKey mapping (old -> new):\n{_km}"
        ) from err

    # May have an explicitly given extent, in which case the space size and
    # offset given by the grid properties needs to be overwritten.
    if extent:
        _l, _r, _b, _t = extent
        space_size = (abs(_r - _l), abs(_t - _b))
        space_offset = (min(_l, _r), min(_b, _t))

    # Check validity
    COORDINATE_MODES = ("offset",)
    if coordinate_mode not in COORDINATE_MODES:
        raise ValueError(
            f"Invalid coordinate mode '{coordinate_mode}'! "
            "Hexagonal grid property `coordinate_mode` needs to be one of:  "
            f"{', '.join(COORDINATE_MODES)}"
        )

    if offset_mode not in ("even", "odd"):
        raise ValueError(
            f"Invalid offset mode '{offset_mode}'! Hexagonal grid property "
            "`offset_mode` needs to be 'even' or 'odd'."
        )

    if boundary not in ("outer", "inner"):
        raise ValueError(
            f"Invalid space boundary '{boundary}'! Hexagonal grid property "
            "`space_boundary` needs to be 'outer' or 'inner'."
        )

    # .. Calculations .........................................................
    # Determine number of cells in x and y direction (of the final plot)
    n_x = data.sizes[x]
    n_y = data.sizes[y]

    ids_x = np.arange(n_x)
    ids_y = np.arange(n_y)

    # Issue a warning if there are too few cells.
    if n_x < 2 or n_y < 2:
        warnings.warn(
            "Plotting a hexagonal grid with fewer than two cells in any "
            "dimension may lead to unexpected results!",
            UserWarning,
        )

    # Decide on which unit hexagon to use
    if pointy_top:
        unit_hexagon = unit_hexagon_pointy_top
    else:
        unit_hexagon = unit_hexagon_flat_top

    # Depending on whether the space is known, need to go different ways:
    #   - If known, deduce s and scale hexagon accordingly.
    #   - If not known, use s = 1 and compute the size of the space instead.
    if space_given:
        # With a space given, the size of the hexagon may be different in the
        # x and y direction in order to cover the whole space with the
        # available number of cells.
        #
        # However, a scaled hexagon no longer has a characteristic size `s`,
        # but needs two parameters to describe the scaling: `lx` and `ly`.
        # These are slightly different than `s`, but depending on which pointy
        # end aligns with which dimension, one of them is equal to what would
        # be `s` in a regular hexagon.
        # To reduce pitfalls of using `s` further down, we deliberately do NOT
        # define it here.

        # First, compute the boundaries (lower, upper) along each dimension
        space_x = (space_offset[0], space_offset[0] + space_size[0])
        space_y = (space_offset[1], space_offset[1] + space_size[1])

        space_width, space_height = space_size

        # Now deduce the scaling factors, depending on pointyness and position
        # of the boundary:
        if pointy_top:
            if boundary == "outer":
                offs_corr = 1 if n_x > 1 else 0  # offset correction
                lx = 2 * space_width / (sqrt_3 * (2 * n_x + offs_corr))
                ly = 2 * space_height / (3 * n_y + 1)
            else:
                lx = 2 * space_width / (sqrt_3 * (2 * n_x - 1))
                ly = 2 * space_height / (3 * n_y - 1)

            cell_width = sqrt_3 * lx  # sic! lx is a scaling factor
            cell_height = 2 * ly

        else:
            # flat top
            if boundary == "outer":
                offs_corr = 1 if n_y > 1 else 0
                lx = 2 * space_width / (3 * n_x + 1)
                ly = 2 * space_height / (sqrt_3 * (2 * n_y + offs_corr))
            else:
                lx = 2 * space_width / (3 * n_x - 1)
                ly = 2 * space_height / (sqrt_3 * (2 * n_y - 1))

            cell_width = 2 * lx
            cell_height = sqrt_3 * ly  # sic! ly is a scaling factor

        # Scale and transform the hexagon accordingly.
        # This may lead to elongation along x or y dimensions.
        # Note the use of lx and ly here, not cell_width and cell_height.
        # This is because lx/ly == 1 if the hexagon is regular (the available
        # space's aspect ratio is 1:sqrt(3)/2 ).
        # In contrast, cell_width / cell_height will be sqrt(3)/2 == 1.15470…
        # for a unit hexagon, thus not suitable as scaling factors.
        hexagon = scale * np.array([lx, ly]) * unit_hexagon

    else:
        # Space was not given, can choose s and deduce space as we desire.
        #
        # Also, the hexagon will be regular: a _uniformly_ scaled unit hexagon.
        # Thus, we do not need to worry about scaling factors like above, but
        # the side length `s` alone suffices.
        s = 1

        hexagon = scale * s * unit_hexagon

        if pointy_top:
            cell_width = s * sqrt_3
            cell_height = 2 * s

            if boundary == "outer":
                offs_corr = 0.5 if n_x > 1 else 0
                space_x = (0.0, (n_x + offs_corr) * cell_width)
                space_y = (0.0, 3 / 2 * n_y * s + s / 2)
            else:
                space_x = (cell_width / 2, n_x * cell_width)
                space_y = (s / 2, 3 / 2 * n_y * s)
        else:
            # flat top
            cell_width = 2 * s
            cell_height = s * sqrt_3

            if boundary == "outer":
                offs_corr = 0.5 if n_y > 1 else 0
                space_x = (0.0, 3 / 2 * n_x * s + s / 2)
                space_y = (0.0, (n_y + offs_corr) * cell_height)
            else:
                space_x = (s / 2, 3 / 2 * n_x * s)
                space_y = (cell_height / 2, n_y * cell_height)

        space_size = (space_x[1], space_y[1])
        space_width, space_height = space_size

        space_x = (space_offset[0] + space_x[0], space_offset[0] + space_x[1])
        space_y = (space_offset[1] + space_y[0], space_offset[1] + space_y[1])

    # .. Compute cell positions . . . . . . . . . . . . . . . . . . . . . . . .
    # Temporary position values -- without row/col offsets!
    if pointy_top:
        if boundary == "outer":
            _pos_x = np.linspace(
                space_x[0] + cell_width / 2,
                space_x[1] - cell_width,  # due to row offset
                n_x,
            )
            _pos_y = np.linspace(
                space_y[0] + cell_height / 2,
                space_y[1] - cell_height / 2,
                n_y,
            )
        else:
            _pos_x = np.linspace(
                space_x[0],
                space_x[1] - cell_width / 2,  # due to row offset
                n_x,
            )
            _pos_y = np.linspace(
                space_y[0] + cell_height / 4,
                space_y[1] - cell_height / 4,
                n_y,
            )
    else:
        # flat top
        if boundary == "outer":
            _pos_x = np.linspace(
                space_x[0] + cell_width / 2,
                space_x[1] - cell_width / 2,
                n_x,
            )
            _pos_y = np.linspace(
                space_y[0] + cell_height / 2,
                space_y[1] - cell_height,  # due to col offset
                n_y,
            )
        else:
            _pos_x = np.linspace(
                space_x[0] + cell_width / 4,
                space_x[1] - cell_width / 4,
                n_x,
            )
            _pos_y = np.linspace(
                space_y[0],
                space_y[1] - cell_height / 2,  # due to col offset
                n_y,
            )

    # Bring into the form that's required for imshow
    x_offsets, y_offsets = np.meshgrid(_pos_x, _pos_y)

    # Add the offset towards higher values
    if pointy_top:
        if n_x > 1:
            offset = cell_width / 2
            if offset_mode == "even":
                x_offsets[ids_y % 2 == 0, ...] += offset
            else:
                x_offsets[ids_y % 2 == 1, ...] += offset
    else:
        if n_y > 1:
            offset = cell_height / 2
            if offset_mode == "even":
                y_offsets[..., ids_x % 2 == 0] += offset
            else:
                y_offsets[..., ids_x % 2 == 1] += offset

    # .. Create the PolyCollection ............................................
    # At this point, need the following information to generate the collection
    # of hexagons:
    #   - the appropriately transformed hexagon
    #   - x and y offsets (2D arrays, flattened and combined)

    collection_kwargs = ensure_dict(collection_kwargs)

    # Here we go ...
    pcoll = mpl.collections.PolyCollection(
        [hexagon],
        offsets=np.transpose([x_offsets.flatten(), y_offsets.flatten()]),
        transOffset=mpl.transforms.AffineDeltaTransform(ax.transData),
        #
        # Pass collection-related kwargs
        linewidths=collection_kwargs.get("linewidths", 0),
        **collection_kwargs,
    )
    # NOTE There also is a RegularPolyCollection, but that is a massive pain
    #      because it expects the sizes to be given in units of the canvas
    #      (area in *points squared*), which depends on the representation and
    #      not on the data.
    #      The PolyCollection does not have that issue because the polygons
    #      need to be drawn "by hand". This way, all information can be
    #      supplied in units of data space when using the data transformation
    #      of the offsets.

    # Set the data (in a consistently flattened form)
    pcoll.set_array(_flatten_hexgrid_data(data))

    # Set cmap stuff, norm, limits
    pcoll.set_cmap(cmap)
    pcoll.set_norm(norm)
    pcoll.set_clim(vmin, vmax)

    # .. Add to axis ..........................................................
    if ax is None:
        import matplotlib.pyplot as plt

        ax = plt.gca()

    ax.add_collection(pcoll)

    # Use same length scale in x and y and set space limits
    ax.set_aspect("equal")
    ax.set_xlim(*space_x)
    ax.set_ylim(*space_y)

    # Allow marking the center points of the hexagons
    #
    # Again need to use a manually created collection such that positions and
    # sizes can be given in data units.
    # The radius factor is a heuristic value for an "effective" `s` parameter
    # for the shorter side of the hexagon.
    if draw_centers:
        circle = mpl.patches.Circle(
            (0, 0),
            radius=min(cell_width, cell_height) / 2 * draw_center_radius,
        )
        draw_center_kwargs = ensure_dict(draw_center_kwargs)
        ccoll = mpl.collections.PatchCollection(
            [circle],
            offsets=np.transpose([x_offsets.flatten(), y_offsets.flatten()]),
            transOffset=mpl.transforms.AffineDeltaTransform(ax.transData),
            #
            linewidths=draw_center_kwargs.pop("linewidths", 0),
            **draw_center_kwargs,
        )
        ax.add_collection(ccoll)

    # If space was not known, don't show axis labels
    if hide_ticks or (not space_given and hide_ticks is None):
        ax.tick_params(
            axis="both",
            left=False,
            top=False,
            right=False,
            bottom=False,
            labelleft=False,
            labeltop=False,
            labelright=False,
            labelbottom=False,
        )

    # Create axes image to have the same result object as imshow does,
    # including interpolation features etc
    im = mpl.image.AxesImage(ax, **im_kwargs)
    im.hexagons = pcoll

    # Do some post-processing
    im.set_cmap(cmap)
    im.set_clim(vmin, vmax)

    # Some callbacks
    # def on_changed(collection):
    #     hbar.set_cmap(collection.get_cmap())
    #     hbar.set_clim(collection.get_clim())
    #     vbar.set_cmap(collection.get_cmap())
    #     vbar.set_clim(collection.get_clim())

    # pcoll.callbacks.connect("changed", on_changed)

    return im


@make_facet_grid_plot(
    map_as="dataarray",
    register_as_kind="imshow_hexagonal",
    encodings=("x", "y"),
    supported_hue_styles=(),
    parse_cmap_and_norm_kwargs=True,
)
def imshow_hexagonal_facet_grid(
    data: xr.DataArray,
    *,
    hlpr: PlotHelper,
    _is_facetgrid: bool,
    x: str = None,
    y: str = None,
    extend: str = None,
    levels: int = None,
    add_labels: bool = True,
    add_colorbar: bool = True,
    cbar_kwargs: dict = None,
    **kwargs,
) -> mpl.image.AxesImage:
    """Wrapper around :py:func:`imshow_hexagonal` that makes it work as a
    standalone, DAG-supporting and *faceting* plotting function.
    Uses :py:class:`~dantro.plot.funcs.generic.make_facet_grid_plot` wrapper.

    For more arguments, see the respective docstrings.

    Args:
        data (xr.DataArray): The to-be-plotted data as prepared by the wrapper.
        hlpr (PlotHelper): The plot helper
        _is_facetgrid (bool): *Internally used variable* that denotes whether
            the invocation is part of a facet grid plot.
        x (str, optional): Which data dimension to represent on the x-axis
        y (str, optional): Which data dimension to represent on the y-axis
        extend (str, optional): Whether to extend the colorbar
        levels (int, optional): Number of discrete colormap levels to use;
            *currently not supported!*
        add_labels (bool, optional): Whether to add labels to the x and y axis.
        add_colorbar (bool, optional): Whether to add a colorbar.
        cbar_kwargs (dict, optional): Colorbar kwargs that are *only used if
            no facet grid* is created.
            **Note:** This interface is subject to change, aim being that the
            arguments can be supplied in the same way for faceting and
            non-faceting invocations of this function.
        **kwargs: Passed on to :py:func:`imshow_hexagonal`
    """
    if levels:
        raise NotImplementedError("`levels` argument not yet supported")

    im = imshow_hexagonal(
        data,
        ax=hlpr.ax,
        x=x,
        y=y,
        **kwargs,
    )

    if not _is_facetgrid:
        if add_labels:
            hlpr.ax.set_xlabel(x)
            hlpr.ax.set_ylabel(y)

        if add_colorbar:
            # TODO This should read information from the FacetGrid's
            #      cbar_kwargs, which are also parsed there... problem being
            #      that there is no FacetGrid object available here.
            #      However, the arguments should be parsed in the same way!
            cbar = hlpr.fig.colorbar(
                im,
                ax=hlpr.ax,
                extend=extend,
                **ensure_dict(cbar_kwargs),
            )

    return im


# .............................................................................


@is_plot_func(use_dag=True, supports_animation=True)
def caplot(
    *,
    hlpr: PlotHelper,
    data: dict,
    to_plot: Dict[str, dict],
    from_dataset: xr.Dataset = None,
    frames: str = "time",
    frames_isel: Union[int, Sequence] = None,
    grid_structure: str = None,
    aspect: float = 1.0,
    aspect_pad: float = 0.1,
    size: float = None,
    col_wrap: Union[int, str, bool] = "auto",
    imshow_hexagonal_extra_kwargs: dict = None,
    default_imshow_kwargs: dict = None,
    default_cbar_kwargs: dict = dict(fraction=0.04, aspect=20),
    suptitle_fstr: str = "{} = {}",
    suptitle_kwargs: dict = None,
):
    """Plots an animated series of one or many 2D Cellular Automata states.

    The data used for plotting is assembled from ``data`` using the keys that
    are specified in ``to_plot``. Alternatively, the ``from_dataset`` argument
    can be used to pass a dataset which contains all the required data.

    The keys in ``to_plot`` should match the names of data variables.
    The values in ``to_plot`` specify the individual subplots' properties like
    the color map that is to be used or the minimum or maximum values.

    For plotting square grids, :py:meth:`matplotlib.axes.Axes.imshow` is used
    and generates output like this:

    .. image:: ../_static/_gen/caplot/snapshot_square.pdf
        :target: ../_static/_gen/caplot/snapshot_square.pdf
        :width: 100%

    For a grid with hexagonal cells, :py:func:`imshow_hexagonal` is used; more
    details on how the cells are mapped to the space can be found there.
    The output (for the same dummy data as used above) may look like this:

    .. image:: ../_static/_gen/caplot/snapshot_hex.pdf
        :target: ../_static/_gen/caplot/snapshot_hex.pdf
        :width: 100%

    Finally, this plot function is specialized to generate animations along the
    ``frames`` dimension of the data, e.g. ``time``:

    .. raw:: html

        <video width="720" src="../_static/_gen/caplot/anim_square.mp4" controls></video>

    .. raw:: html

        <video width="720" src="../_static/_gen/caplot/anim_hex.mp4" controls></video>

    .. admonition:: Requirements on the CA data

        The selected data (keys in ``to_plot`` that correspond to DAG results
        in ``data``) should have two *spatial* dimensions and one data
        dimension that goes along the ``frames`` dimension.

        All coordinates should be identical, otherwise the behavior is not
        defined or alignment might fail.

        **For hexagonal grid structure**, note the requirements given in
        :py:func:`imshow_hexagonal`.

    .. admonition:: See also

        * :ref:`plot_funcs_ca`
        * :ref:`plot_funcs_ca_hex`

    Args:
        hlpr (PlotHelper): The plot helper instance
        data (dict): The selected data
        to_plot (Dict[str, dict]): Which data to plot and how. The keys of
            this dict refer to an item within the selected ``data`` or the
            given dataset.
            Each of these keys is expected to hold yet another dict,
            supporting the following configuration options (all optional):

            - ``title`` (str, optional):
                The title for this sub-plot.
            - ``cmap`` (Union[str, list, dict], optional):
                Which colormap to use. This argument is handled by the
                :py:class:`~dantro.plot.utils.color_mngr.ColorManager`,
                providing many ways in which to define the colormap.
                For instance, by passing mapping from labels to colors, a
                discrete colormap is created: The keys will be the labels and
                the values will be their colors. Association happens in the
                order of entries, with values being inferred from ``vmin`` and
                ``vmax``, if given.
                For more information and examples, see the docstring of the
                :py:class:`~dantro.plot.utils.color_mngr.ColorManager`.
            - ``norm`` (Union[str, dict], optional):
                The normalization function to use, also handled by the
                :py:class:`~dantro.plot.utils.color_mngr.ColorManager`.
            - ``vmin`` (float, optional):
                The fixed lower data limit for this property; if not given,
                uses auto-scaling, which may lead to jumps in the animation.
                Can also be ``'min'`` in which case the *global* minimum of the
                available data is used.
            - ``vmax`` (float, optional):
                Same as ``vmin``, but with the maximum and allowing ``'max'``
                argument for choosing the global maximum.
            - ``label`` (str, optional):
                The *colorbar* label.
            - ``imshow_kwargs`` (dict, optional):
                Passed on to the imshow invocation, i.e. to
                :py:meth:`~matplotlib.axes.Axes.imshow` or
                :py:meth:`imshow_hexagonal`.
            - ``**kwargs``:
                Further arguments control the colorbar appearance, their
                labels, ticks, and other plot specifics. For more detailed
                information about available arguments, see
                :py:func:`._plot_ca_property`, which takes care of plotting
                the individual ``to_plot`` entries.

        from_dataset (xarray.Dataset, optional): If given, will use this object
            instead of assembling a dataset from ``data`` and ``to_plot`` keys.
        frames (str, optional): Name of the animated dimension, typically the
            time dimension.
        frames_isel (Union[int, Sequence], optional): The index selector for
            the frames dimension. Can be a single integer but also a range
            expression.
        grid_structure (str, optional): The underlying grid structure, can be
            ``square``, ``hexagonal``, or ``triangular`` (not implemented).
            If None, will try to read it from the individual properties' data
            attribute ``grid_structure``.
        aspect (float, optional): The aspect ratio (width/height) of the
            subplots; should match the aspect ratio of the data.
        aspect_pad (float, optional): A factor that is added to the calcuation
            of the subplots width. This can be used to create a horizontal
            padding between subplots.
        size (float, optional): The height in inches of a subplot.
            Is used to determine the subplot size with the width being
            calculated by ``size * (aspect + aspect_pad)``.
        col_wrap (Union[int, str, bool], optional): Controls column wrapping.
            If ``auto``, will compute a column wrapping that leads to an
            (approximately) square subplots layout (not taking into account
            the subplots aspect ratio, only the grid layout). This will start
            producing a column wrapping with four or more properties to plot.
        default_imshow_kwargs (dict, optional): The default parameters passed
            to the underlying imshow plotting function. These are updated by
            the values given via ``to_plot``.
        default_cbar_kwargs (dict, optional): The default parameters for the
            colorbar that is added to applicable subplots. These are updated
            by the parameters given under ``to_plot``.
        suptitle_fstr (str, optional): A format string used to create the
            suptitle string. Passed arguments are ``frames`` and the currently
            selected frames *coordinate* (not the index!).
            If this evaluates to False, will not create a figure suptitle.
        suptitle_kwargs (dict, optional): Passed on to ``fig.suptitle``.
    """
    # Helper functions ........................................................

    def get_grid_structure(d: xr.DataArray) -> str:
        """Retrieves the grid structure from data attributes"""
        grid_structure = d.attrs.get("grid_structure")
        if isinstance(grid_structure, np.ndarray):
            grid_structure = grid_structure.item()

        return grid_structure

    def prepare_data(data: dict, *, prop_name: str) -> xr.DataArray:
        """Prepares data for (later) creating an :py:class:`xarray.Dataset`"""
        d = data[prop_name]
        if isinstance(d, AbstractDataContainer):
            d = d.data
        return d

    def select_data(ds: xr.Dataset, name: str, isel: dict) -> xr.DataArray:
        """Selects a slice of data for plotting using
        :py:meth:`~xarray.DataArray.isel` on the data variable ``name``."""
        return ds[name].isel(isel)

    def set_suptitle(data: xr.DataArray):
        """Sets the suptitle"""
        if not suptitle_fstr:
            return

        hlpr.fig.suptitle(
            suptitle_fstr.format(frames, data.coords[frames].item()),
            **(suptitle_kwargs if suptitle_kwargs else {}),
        )

    # Prepare the data ........................................................
    # Work on a copy of the plot spec
    to_plot = copy.deepcopy(to_plot)

    # Bring data into xr.Dataset form
    log.note("Preparing data for CA plot ...")
    if from_dataset:
        log.remark("Using explicitly passed dataset ...")
        ds = from_dataset
    else:
        log.remark("Constructing dataset ...")
        ds = xr.Dataset({p: prepare_data(data, prop_name=p) for p in to_plot})

    # Check that frames dimension is available
    if not frames or frames not in ds.coords:
        _avail = ", ".join(ds.coords.keys())
        raise ValueError(
            f"Invalid `frames` coordinate dimension '{frames}'! "
            f"Available coordinates:  {_avail}"
        )

    # Apply selection along frames dimension
    if frames_isel is not None:
        if isinstance(frames_isel, int):
            frames_isel = [frames_isel]
        _selector = {frames: frames_isel}
        log.remark("Applying index selection %s  ...", _selector)
        ds = ds.isel(_selector, drop=False)

    # TODO x-y mapping?
    # TODO Automate aspect computation from x and y dimensions

    # Depending on length of coordinate dimension, ensure that animation mode
    # is enabled or disabled.
    num_frames = ds.coords[frames].size
    if num_frames > 1:
        hlpr.enable_animation()
    else:
        hlpr.disable_animation()

    # If not given, retrieve the structure from the data variable's attributes.
    if not grid_structure:
        structures = {p: get_grid_structure(ds[p]) for p in to_plot}

        if len(set(structures.values())) != 1:
            raise ValueError(
                "Mismatch in grid structure; all grid structures need to be "
                f"the same but were:  {structures}\n"
                "This may have resulted from data attributes being lost in a "
                "data transformation. If so, one alternative to re-adding the "
                "data attributes (via the `update_with_attrs_from` operation) "
                "is to specify `grid_structure` explicitly. "
                "For hexagonal grids, grid properties can also be passed via "
                "`imshow_kwargs.grid_properties`."
            )
        grid_structure = next(iter(structures.values()))

    # Evaluate limits argument for all properties
    for prop_name, spec in to_plot.items():
        if spec.get("vmin") == "min":
            spec["vmin"] = ds[prop_name].min().item()

        if spec.get("vmax") == "max":
            spec["vmax"] = ds[prop_name].max().item()

    # Inform about the data that is to be plotted
    log.note(
        "Performing CA plot for %d data variable%s ...",
        len(to_plot),
        "" if len(to_plot) == 1 else "s",
    )
    log.remark("  Data variables:     %s", ", ".join(to_plot))
    log.remark(
        "  Dimensions:         %s",
        ", ".join(f"{k}: {s}" for k, s in ds.sizes.items()),
    )
    log.remark("  Grid structure:     %s", grid_structure)

    # Some final checks ...
    if len(ds.sizes) != 3:
        raise ValueError(
            "Dataset shape needs to be 3-dimensional, but was: "
            f"{dict(ds.sizes)}! Full dataset:\n{ds}"
        )

    # Prepare the figure ......................................................
    # Evaluate column wrapping
    if col_wrap and not (col_wrap == "auto" and len(to_plot) < 4):
        if col_wrap == "auto":
            col_wrap = ceil(sqrt(len(to_plot)))
        log.remark("  Column wrapping:    %s", col_wrap)

        ncols = col_wrap
        nrows = ceil(len(to_plot) / col_wrap)
        axis_map = {
            p: dict(col=i % col_wrap, row=i // col_wrap)
            for i, p in enumerate(to_plot)
        }

    else:
        ncols = len(to_plot)
        nrows = 1
        axis_map = {p: dict(col=col, row=0) for col, p in enumerate(to_plot)}

    # Determine the figsize from the size argument (height) and data aspect.
    if not size:
        size = mpl.rcParams["figure.figsize"][1]
    figsize = (size * (aspect + aspect_pad), size)

    # Create the figure and set all axes as invisible. This is needed because
    # col_wrap may lead to some subplots being completely empty.
    hlpr.setup_figure(
        figsize=figsize,
        scale_figsize_with_subplots_shape=True,
        ncols=ncols,
        nrows=nrows,
    )
    for ax in hlpr.axes.flat:
        ax.set_visible(False)

    # Do the single plot for all data variables, looping through subfigures.
    # This creates the first imshow objects, which are being kept track of
    # such that they can later be updated.
    # This also sets the axes as visible again, but only for those that have
    # a property assigned to them.
    ims = dict()
    for i, (prop_name, props) in enumerate(to_plot.items()):
        hlpr.select_axis(**axis_map[prop_name])
        hlpr.ax.set_visible(True)

        # Select the appropriate data, then plot the data variable
        data = select_data(ds, prop_name, {frames: 0})

        ims[prop_name] = _plot_ca_property(
            prop_name,
            hlpr=hlpr,
            data=data,
            grid_structure=grid_structure,
            default_imshow_kwargs=default_imshow_kwargs,
            imshow_hexagonal_extra_kwargs=imshow_hexagonal_extra_kwargs,
            default_cbar_kwargs=default_cbar_kwargs,
            **props,
        )

        # Use this data for setting the figure suptitle
        if i == 0:
            set_suptitle(data)

    # End of single frame CA state plot function ..............................
    # The above variables are all available below, but the update function is
    # supposed to start plotting anew, starting from frame 0.

    def update_data():
        """Updates the data of the imshow objects"""

        def need_autoscale(*, vmin=None, vmax=None, cmap=None, **_) -> bool:
            """Returns True if there are no bounds (vmin, vmax) given
            or if a discrete colormap is created from a dict"""
            return vmin is None and vmax is None and not isinstance(cmap, dict)

        log.note("Plotting animation with %d frames ...", num_frames)

        # Determine whether a autoscaling is needed for a property:
        needs_autoscale = {
            n: need_autoscale(**spec) for n, spec in to_plot.items()
        }

        # Frame iteration
        for frame_idx in range(num_frames):
            log.debug("Plotting frame %d ...", frame_idx)

            for i, (prop_name, props) in enumerate(to_plot.items()):
                hlpr.select_axis(**axis_map[prop_name])

                frame_data = select_data(ds, prop_name, {frames: frame_idx})

                # Depending on grid structure, update data or plot anew
                if grid_structure == "hexagonal":
                    # Update imshow_hexagonal data according as would otherwise
                    # happen inside that function
                    frame_data, _, _ = _prepare_hexgrid_data(
                        frame_data, x=props.get("x"), y=props.get("y")
                    )
                    ims[prop_name].hexagons.set_array(
                        _flatten_hexgrid_data(frame_data)
                    )

                    if needs_autoscale[prop_name]:
                        ims[prop_name].hexagons.autoscale()

                else:
                    # Update imshow data without creating a new object
                    ims[prop_name].set_data(frame_data.T)

                    if needs_autoscale[prop_name]:
                        ims[prop_name].autoscale()

                # Use the first subplot's data for setting the figure suptitle
                if i == 0:
                    set_suptitle(frame_data)

            # Done with this frame; yield control to the animation framework.
            yield

    # Register this update method with the helper, which takes care of the rest
    hlpr.register_animation_update(update_data)
